# -*- coding: utf-8 -*-

# modified hierachical.py from scikit learn
"""Hierarchical Agglomerative Clustering

These routines perform some hierarchical agglomerative clustering of some input data. Currently, only Ward's algorithm
is implemented.

Authors : Vincent Michel, Bertrand Thirion, Alexandre Gramfort, Gael Varoquaux
Modified: Aina Frau
License: BSD 3 clause

"""

import warnings
import logging

from heapq import heapify, heappop, heappush, heappushpop

import numpy as np

from scipy import sparse
from scipy.cluster import hierarchy
try:
    from scipy.maxentropy import logsumexp
except ImportError:
    from scipy.misc import logsumexp
from scipy.sparse import cs_graph_components
try:
    from sklearn.base import BaseEstimator, ClusterMixin
    from sklearn.externals.joblib import Memory
    from sklearn.metrics import euclidean_distances
    from sklearn.utils import array2d
    from sklearn.cluster._feature_agglomeration import AgglomerationTransform
    from sklearn.mixture import GMM
    from sklearn.cluster import Ward as WardSK
except ImportError:
    class BaseEstimator(object):
        pass
    class ClusterMixin(object):
        pass
    def Memory(*args, **kwargs):
        pass
    class AgglomerationTransform(object):
        pass
    # This is only for tests including doctests (since nose try to import
    # everything) in the case of minimal requirements (without scikit-learn)
    # TODO: find a better way of handling this case (or make sklearn a required
    # dependency)

import pyhrf

from pyhrf.tools import add_prefix


logger = logging.getLogger(__name__)


# Functions for Ward algorithm
def loglikelihood_computation(fm, mu0, v0, mu1, v1, a):
    # log likelihood computation
    r2 = np.logaddexp(np.log(a / v1 ** (n / 2.)) - (fm - mu1) ** 2 / (2 * v1),
                      np.log((1 - a) / v0 ** (n / 2.)) - (fm - mu0) ** 2 / (2 * v0)).sum()
    return r2


def informedGMM(features, alphas):
    """
    Given a set of features, parameters (mu, v, lambda), and alphas:
    updates the parameters
    WARNING: only works for nb features = 1
    """
    if len(features.shape) > 1:
        n = features.shape[1]
        fm = features
    else:
        n = 1
        fm = features[:, np.newaxis]
    am = alphas
    eps = 1e-6
    # parameters estimates
    if am.sum(0) > 0:
        mu1 = (fm.T * am).T.sum(0) / am.sum(0)
        v1 = (((fm - mu1) ** 2).T * am).T.sum(0) / (n * am.sum(0))
        v1 = max(eps, v1)
    else:
        mu1 = v1 = eps
    if (1 - am).sum(0) > 0:
        mu0 = (fm.T * (1 - am)).T.sum(0) / (1 - am).sum(0)
        v0 = (((fm - mu0) ** 2).T * (1 - am)).T.sum(0) / (n * (1 - am).sum(0))
        v0 = max(eps, v0)
    else:
        mu0 = v0 = eps
    a = am.mean()
    if 0:
        print 'Updated parameters'
        print 'mu1:', mu1, 'mu0:', mu0
        print 'v1:', v1, 'v0:', v0
        print 'lambda:', 1 - a
        print ''

    return np.array([mu0, mu1]).squeeze(), np.array([v0, v1]).squeeze(), \
        np.array([1 - a, a]).squeeze()


def norm2_bc(a, b):
    """
    broadcast the computation of ||a-b||^2
    where size(a) = (m,n), size(b) = n
    """
    return (a ** 2).sum(1) + (b ** 2).sum() - 2 * (a * b).sum(1)


def informedGMM_MV(fm, am, cov_type='spherical'):
    """
    Given a set of multivariate features, parameters (mu, v, lambda), and alphas:
    fit a GMM where posterior weights are known (alphas)
    """
    n = fm.shape[1]
    eps = 1e-6
    #eps_ = np.zeros(n) + eps

    # parameters estimates
    if am.sum(0) > 0:
        mu1 = (fm.T * am).T.sum(0) / am.sum(0)
        #v1 = np.dot(np.dot((fm - mu1),(fm - mu1).T).T , am).T.sum(0) / (am.sum()*n)
        v1 = 0.0
        for ns in xrange(0, fm.shape[0]):
            v1 += (fm[ns, :] - mu1) * (fm[ns, :] - mu1).T * am[ns]
        v1 = v1 / (am.sum() * n)
        for nf in xrange(0, len(v1)):
            v1[nf] = max(eps, v1[nf])
        # old spherical covariance:
        #v1b = (norm2_bc(fm, mu1).T * am).T.sum(0) / (am.sum()*n)
    else:
        mu1 = fm.mean(0)
        v1m = fm.var()
        v1 = np.array([v1m, v1m])

    if (1 - am).sum(0) > 0:
        mu0 = (fm.T * (1 - am)).T.sum(0) / (1 - am).sum()
        #v0 = np.dot(np.dot((fm - mu0),(fm - mu0).T).T , am).T.sum(0) / (am.sum()*n)
        v0 = 0.0
        for ns in xrange(0, fm.shape[0]):
            v0 += (fm[ns, :] - mu1) * (fm[ns, :] - mu1).T * (1 - am)[ns]
        v0 = v0 / ((1 - am).sum() * n)
        for nf in xrange(0, len(v0)):
            v0[nf] = max(eps, v0[nf])
        # old spherical covariance:
        #v0b = (norm2_bc(fm, mu0).T * (1-am)).T.sum(0) / ((1-am).sum()*n)
    else:
        mu0 = fm.mean(0)
        v0m = fm.var()
        v0 = np.array([v0m, v0m])

    a = am.mean()

    if cov_type == 'spherical':
        v0m = np.mean(v0)
        v0 = np.array([v0m, v0m])
        v1m = np.mean(v1)
        v1 = np.array([v1m, v1m])

    if 0:
        print 'mu1:', mu1, 'mu0:', mu0
        print 'v1:', v1, 'v0:', v0
        print 'lambda:', a
        print ''

    # log likelihood computation
    aux1 = 0.0
    for ns in xrange(0, fm.shape[0]):
        # /n
        aux1 += np.dot(np.dot((fm[ns, :] - mu1),
                              np.linalg.inv(np.diag(v1))), (fm[ns, :] - mu1).T)
    aux0 = 0.0
    for ns in xrange(0, fm.shape[0]):
        # /n
        aux0 += np.dot(np.dot((fm[ns, :] - mu0),
                              np.linalg.inv(np.diag(v0))), (fm[ns, :] - mu0).T)
    loglh0 = np.logaddexp(np.log(a / (np.linalg.det(np.diag(v1)) * 2 * np.pi) ** (n / 2.)) - aux1 / 2,
                          np.log((1 - a) / (np.linalg.det(np.diag(v0)) * 2 * np.pi) ** (n / 2.)) - aux0 / 2)
    # old spherical log-likelihood
    # loglh0b = np.logaddexp(np.log(a/(v1[0]*2*np.pi)**(n/2.)) - norm2_bc(fm,mu1)/(2*v1[0]),
    # np.log((1-a)/(v0[0]*2*np.pi)**(n/2.)) - norm2_bc(fm,mu0)/(2*v0[0]))
    loglh = np.log(np.sum(np.exp(loglh0)))  # logsumexp(lpr)

    if np.isnan(loglh):
        print 'mu1:', mu1, 'mu0:', mu0
        print 'v1:', v1, 'v0:', v0
        print 'lambda:', a
        print ''
        raise Exception('log-likelihood is nan')

    return mu0, mu1, v0, v1, a, loglh

"""
def informedGMM_MV_old(fm, am, cov_type = 'spherical'):

    n = fm.shape[1]
    eps =  1e-6
    a = am.mean()
    #eps_ = np.zeros(n) + eps
    print 'n_features = ', n
    print 'n_samples = ', fm.shape[0]
    # parameters estimates
    if am.sum(0) > 0:
        mu1 = (fm.T * am).T.sum(0) / am.sum(0)
        #cov1 = np.dot(a.T, fm*fm)/am - 2*mu1*fm/am + mu1**2 # covar a*||fm-mu||^2/am
        #lpr = -0.5 * ( n * np.log(2*np.pi) + np.sum(np.log(cov1),1) + \
        #               np.sum((mu1**2)/cov1,1) - 2 * np.dot(fm, (mu1/cov1).T) + \
        #               np.dot(fm**2,(1.0/cov1).T))
        if cov_type == 'diag':
            v1 = np.dot(np.dot((fm - mu1),(fm - mu1).T).T , am).T.sum(0) / (am.sum()*n)
            if v1 == 0.0:
                print 'zero'
                v1 = np.ones(n)*eps
                print v1
                print v1.shape
            else:
                print 'max'
                for nf in xrange(0, len(v1)):
                    v1[nf] = max(eps,v1[nf])
        else:        # cov_type == 'spherical'
            v1b = (norm2_bc(fm, mu1).T * am).T.sum(0) / (am.sum()*n)
            v1b = max(eps, v1b)
            v1 = np.array([v1b, v1b])
    else:
        mu1 = fm.mean(0)
        v1b = fm.var()
        v1 = np.array([v1b, v1b])
    if (1-am).sum(0) > 0:
        mu0 = (fm.T * (1-am)).T.sum(0) / (1-am).sum()
        if cov_type == 'diag':
            v0 = np.dot(np.dot((fm - mu0),(fm - mu0).T).T , am).T.sum(0) / (am.sum()*n)
            for nf in xrange(0, len(v0)):
                v0[nf] = max(eps,v0[nf])
        else:       # cov_type == 'spherical'
            v0b = (norm2_bc(fm, mu0).T * (1-am)).T.sum(0) / ((1-am).sum()*n)
            v0b = max(eps, v0b)
            v0 = np.array([v0b, v0b])
    else:
        mu0 = fm.mean(0)
        v0b = fm.var()
        v0 = np.array([v0b, v0b])

    if 1:
        print 'mu1:', mu1, 'mu0:', mu0
        print 'v1:', v1, 'v0:', v0, ', ', v0b
        print 'lambda:', a
        print ''

    # log likelihood computation
    #loglh = np.logaddexp(np.log(a/(v1*2*np.pi)**(n/2.)) - norm2_bc(fm,mu1)/(2*v1),
    #                     np.log((1-a)/(v0*2*np.pi)**(n/2.)) - norm2_bc(fm,mu0)/(2*v0)).sum()
    if 1: #cov_type == 'diag':
        loglh0b = np.logaddexp(np.log(a/(v1[0]*2*np.pi)**(n/2.)) - norm2_bc(fm,mu1)/(2*v1[0]),
                              np.log((1-a)/(v0[0]*2*np.pi)**(n/2.)) - norm2_bc(fm,mu0)/(2*v0[0]))
    #loglh = np.log(np.sum(np.exp(loglh0)))
    loglhb = np.log(np.sum(np.exp(loglh0b))) #logsumexp(loglh0, axis = 1)

    if np.isnan(loglhb):
                print 'mu1:', mu1, 'mu0:', mu0
                print 'v1:', v1, 'v0:', v0
                print 'lambda:', a
                print ''
                raise Exception('log-likelihood is nan')

    return mu0, mu1, v0, v1, a, loglhb
"""


def compute_mixt_dist(features, alphas, coord_row, coord_col, cluster_masks, moments, cov_type, res):
    r"""Within one given territory: bi-Gaussian mixture model with known posterior weights:

    .. math::
        phi ~ \sum_i \lambda_i N(mu_i, v_i)
        p(q_j = i | phi_j) \text{is an input (alphas)}

    **Estimation:**
    :math:`\lambda_1 = 1 - \lambda_0` is the mean of posterior weights. :math:`mu_i` is estimated by weighted sample
    mean and :math:`v_i` is estimated by weighted sample variance.

    Parameters
    ----------
    features : np.array((nsamples, nfeatures), float)
        the feature to parcellate
    alphas : np.array(nsamples, float)
        confidence levels on the features -> identified to posterior weights of class activating in the GMM fit
    coord_row : list of int
        row candidates for merging
    coord_col : list of int
        col candidates for merging
    cluster_masks
    moments
    res

    """
    size_max = coord_row.shape[0]

    def mixt_dist(mom1, mom2, merge_mask, cov_type):
        def log_phy(c, cov_type):
            """
            c : cluster mask
            """

            # print c
            m = np.where(c)
            # print m
            fm = features[m[0], :]
            am = alphas[m]
            #mu0, mu1, v0, v1, a, loglh = informedGMM_MV(fm, am)
            _, _, _, _, _, loglh = informedGMM_MV(fm, am, cov_type)

            return loglh

        #merge_mask = np.bitwise_or(c0, c1)
        # + merge_mask.sum() is used to penalize too big clusters
        # return log_phy(c0) + log_phy(c1) - log_phy(merge_mask) #+ .7 *
        # merge_mask.sum()
        return (mom1 + mom2 - log_phy(merge_mask, cov_type)) + merge_mask.sum()

    for i in range(size_max):
        row = coord_row[i]
        col = coord_col[i]
        #res[i] = mixt_dist(cluster_masks[row], cluster_masks[col])
        cmr, cmc = cluster_masks[row], cluster_masks[col]
        merge_mask = np.bitwise_or(cmr, cmc)
        if merge_mask.sum() == 2:  # weighted euclidian distance
            res[i] = (
                (features[cmr[0], :] - features[cmc[0], :]) ** 2).sum() ** .5
        else:
            res[i] = mixt_dist(
                moments[row], moments[col], merge_mask, cov_type)
        logger.info('dist %d <-> %d = %f', row, col, res[i])
    return res




def compute_mixt_dist_skgmm(features, alphas, coord_row, coord_col,
                            cluster_masks, moments, cov_type, res):
    size_max = coord_row.shape[0]

    def mixt_dist(mom1, mom2, merge_mask, cov_type):
        def log_phy(c, cov_type):
            """
            c : cluster mask
            """
            m = np.where(c)
            if len(m[0]) == 1:  # singleton
                return 0
            else:
                fm = features[m[0], :]
                #n = features.shape[1]
                am = alphas[m]
                #g = GMM(n_states=2, cvtype='diag')
                g = GMM(n_components=2, covariance_type=cov_type)
                return g.fit(fm).score(fm).sum()

        if merge_mask.sum() == 2:  # euclidian distance
            return ((mom1 - mom2) ** 2).sum() ** .5
        else:
            return mom1 + mom2 - log_phy(merge_mask, cov_type) + merge_mask.sum()

    for i in range(size_max):
        row = coord_row[i]
        col = coord_col[i]
        merge_mask = np.bitwise_or(cluster_masks[row], cluster_masks[col])
        res[i] = mixt_dist(moments[row], moments[col], merge_mask, cov_type)
        # print 'row %d, col %d -> %f' %(row, col, res[i])
    return res


# def compute_ward_dist_sk(np.ndarray[DOUBLE, ndim=1, mode='c'] m_1,
#                       np.ndarray[DOUBLE, ndim=2, mode='c'] m_2,
#                       np.ndarray[INTP, ndim=1, mode='c'] coord_row,
#                       np.ndarray[INTP, ndim=1, mode='c'] coord_col,
#                       np.ndarray[DOUBLE, ndim=1, mode='c'] res):
#     cdef INTP size_max = coord_row.shape[0]
#     cdef INTP n_features = m_2.shape[1]
#     cdef INTP i, j, row, col
#     cdef DOUBLE pa, n

#     for i in range(size_max):
#         row = coord_row[i]
#         col = coord_col[i]
#         n = (m_1[row] * m_1[col]) / (m_1[row] + m_1[col])
#         pa = 0.
#         for j in range(n_features):
#             pa += (m_2[row, j] / m_1[row] - m_2[col, j] / m_1[col]) ** 2
#         res[i] = pa * n
#     return res


def compute_uward_dist(m_1, m_2, coord_row, coord_col, variance, actlev, res):
    """ Function computing Ward distance:
    inertia = !!!!0

    Parameters
    ----------
    m_1,m_2,coord_row,coord_col: clusters' parameters
    variance: uncertainty
    actlev: activation level

    Returns
    -------
    res: Ward distance

    Modified: Aina Frau
    """
    size_max = coord_row.shape[0]
    n_features = m_2.shape[1]
    logger.info(
        'ward dist with uncertainty -> %d computations to do', size_max)

    for i in range(size_max):
        row = coord_row[i]
        col = coord_col[i]
        n = (m_1[row] * m_1[col]) / (m_1[row] + m_1[col])  # size cluster
        pa = 0.
        for j in range(n_features):
            pa += (m_2[row, j] - m_2[col, j]) ** 2
        logger.info('pa  %d <-> %d = %f', row, col, pa)
        aux = actlev[row] * actlev[col]
        logger.info('aux %d <-> %d = %f', row, col, aux)
        res[i] = pa * aux  # * n
        logger.info('dist %d <-> %d = %f', row, col, res[i])
        if np.isnan(res[i]):
            raise Exception('inertia is nan')

    return res


def compute_uward_dist2(m_1, features, alphas, coord_row, coord_col, cluster_masks, res):
    """ Function computing Ward distance:
    In this case we are using the model-based definition to compute the inertia

    Parameters
    ----------
    m_1,m_2,coord_row,coord_col: clusters' parameters
    variance: uncertainty
    actlev: activation level

    Returns
    -------
    res: Ward distance

    Modified: Aina Frau
    """
    size_max = coord_row.shape[0]
    n_features = features.shape[1]
    logger.info(
        'ward dist with uncertainty -> %d computations to do', size_max)

    def ward_dist(m0, m1, merge_mask):
        def ess(c):
            """
            c : cluster mask
            """
            m = np.where(merge_mask)
            if len(m[0]) == 1:  # singleton
                return 0
            else:
                fm = features[m[0], :]
                n = features.shape[1]
                am = alphas[m]
                # parameters estimates
                mu0 = fm.mean()
                inertia = (am * ((fm - mu0) ** 2).T).T.sum()
            return inertia

        if merge_mask.sum() == 2:  # euclidian distance
            return ((m0 - m1) ** 2).sum() ** .5
        else:
            return ess(merge_mask) - m0 - m1  # + .7 * merge_mask.sum()

    for i in range(size_max):
        row = coord_row[i]
        col = coord_col[i]
        merge_mask = np.bitwise_or(cluster_masks[row], cluster_masks[col])
        res[i] = ward_dist(m_1[row], m_1[col], merge_mask)
        logger.info('dist %d <-> %d = %f', row, col, res[i])

    return res


def _hc_get_descendent(node, children, n_leaves):
    """
    Function returning all the descendent leaves of a set of nodes in the tree.

    Parameters
    ----------
    node : int
    The node for which we want the descendents.

    children : list of pairs. Length of n_nodes
    The children of each non-leaf node. Values less than `n_samples` refer
    to leaves of the tree. A greater value `i` indicates a node with
    children `children[i - n_samples]`.

    n_leaves : int
    Number of leaves.

    Returns
    -------
    descendent : list of int
    """
    ind = [node]
    if node < n_leaves:
        return ind
    descendent = []
    # It is actually faster to do the accounting of the number of
    # elements is the list ourselves: len is a lengthy operation on a
    # chained list
    n_indices = 1
    while n_indices:
        i = ind.pop()
        if i < n_leaves:
            descendent.append(i)
            n_indices -= 1
        else:
            ind.extend(children[i - n_leaves])
            n_indices += 1
    return descendent


def hc_get_heads(parents, copy=True):
    """ Return the heads of the forest, as defined by parents
    Parameters
    ===========
    parents: array of integers
    The parent structure defining the forest (ensemble of trees)
    copy: boolean
    If copy is False, the input 'parents' array is modified inplace

    Returns
    =======
    heads: array of integers of same shape as parents
    The indices in the 'parents' of the tree heads
    """
    if copy:
        parents = np.copy(parents)
    size = parents.size
    for node0 in range(size):
        # Start from the top of the tree and go down
        node0 = size - node0 - 1
        node = node0
        parent = parents[node]
        while parent != node:
            parents[node0] = parent
            node = parent
            parent = parents[node]
    return parents


def _get_parents(nodes, heads, parents, not_visited):
    """ Return the heads of the given nodes, as defined by parents
    Modifies in-place 'heads' and 'not_visited'

    Parameters
    ===========
    nodes: list of integers
    The nodes to start from
    heads: list of integers
    A list to hold the results (modified inplace)
    parents: array of integers
    The parent structure defining the tree
    not_visited:
    The tree nodes to consider (modified inplace)
    """
    for node in nodes:
        parent = parents[node]
        while parent != node:
            node = parent
            parent = parents[node]
        if not_visited[node]:
            not_visited[node] = 0
            heads.append(node)
    return heads


###############################################################################
# Ward's algorithm

def ward_tree(X, connectivity=None, n_components=None, copy=True, n_clusters=None,
              var=None, act=None, var_ini=None, act_ini=None,
              dist_type='uward', cov_type='spherical', save_history=False):
    """Ward clustering based on a Feature matrix.

    The inertia matrix uses a Heapq-based representation.

    This is the structured version, that takes into account a some topological
    structure between samples.

    Parameters
    ----------
    X : array of shape (n_samples, n_features)
        feature matrix  representing n_samples samples to be clustered

    connectivity : sparse matrix.
        connectivity matrix. Defines for each sample the neigbhoring samples
        following a given structure of the data. The matrix is assumed to
        be symmetric and only the upper triangular half is used.
        Default is None, i.e, the Ward algorithm is unstructured.

    n_components : int (optional)
        Number of connected components. If None the number of connected
        components is estimated from the connectivity matrix.

    copy : bool (optional)
        Make a copy of connectivity or work inplace. If connectivity
        is not of LIL type there will be a copy in any case.

    n_clusters : int (optional)
        Stop early the construction of the tree at n_clusters. This is
        useful to decrease computation time if the number of clusters is
        not small compared to the number of samples. In this case, the
        complete tree is not computed, thus the 'children' output is of
        limited use, and the 'parents' output should rather be used.
        This option is valid only when specifying a connectivity matrix.

    dist_type : str -> uward | mixt

    Returns
    -------
    children : 2D array, shape (n_nodes, 2)
        list of the children of each nodes.
        Leaves of the tree have empty list of children.

    n_components : sparse matrix.
        The number of connected components in the graph.

    n_leaves : int
        The number of leaves in the tree

    parents : 1D array, shape (n_nodes, ) or None
        The parent of each node. Only returned when a connectivity matrix
        is specified, elsewhere 'None' is returned.

    Modified: Aina Frau
    """
    X = np.asarray(X)
    if X.ndim == 1:
        X = np.reshape(X, (-1, 1))
    n_samples, n_features = X.shape

    if connectivity is None:
        if n_clusters is not None:
            warnings.warn('Early stopping is implemented only for '
                          'structured Ward clustering (i.e. with '
                          'explicit connectivity.', stacklevel=2)
        out = hierarchy.ward(X)
        children_ = out[:, :2].astype(np.int)
        return children_, 1, n_samples, None, None, None

    # Compute the number of nodes
    if n_components is None:
        n_components, labels = cs_graph_components(connectivity)

    # Convert connectivity matrix to LIL with a copy if needed
    if sparse.isspmatrix_lil(connectivity) and copy:
        connectivity = connectivity.copy()
    elif not sparse.isspmatrix(connectivity):
        connectivity = sparse.lil_matrix(connectivity)
    else:
        connectivity = connectivity.tolil()

    if n_components > 1:
        warnings.warn("the number of connected components of the "
                      "connectivity matrix is %d > 1. Completing it to avoid "
                      "stopping the tree early." % n_components)
        connectivity = _fix_connectivity(X, connectivity, n_components, labels)
        n_components = 1

    if n_clusters is None:
        n_nodes = 2 * n_samples - n_components
    else:
        assert n_clusters <= n_samples
        n_nodes = 2 * n_samples - n_clusters

    if (connectivity.shape[0] != n_samples
            or connectivity.shape[1] != n_samples):
        raise ValueError('Wrong shape for connectivity matrix: %s '
                         'when X is %s' % (connectivity.shape, X.shape))

    # create inertia matrix
    coord_row = []
    coord_col = []
    A = []
    for ind, row in enumerate(connectivity.rows):
        A.append(row)
        # We keep only the upper triangular for the moments
        # Generator expressions are faster than arrays on the following
        row = [i for i in row if i < ind]
        coord_row.extend(len(row) * [ind, ])
        coord_col.extend(row)

    coord_row = np.array(coord_row, dtype=np.int)
    coord_col = np.array(coord_col, dtype=np.int)

    # compute initial cluster masks
    dummy_idx = np.arange(n_samples)
    cluster_masks = [dummy_idx == i for i in range(n_nodes)]

    inertia = np.empty(len(coord_row), dtype=np.float)
    variance = np.zeros((n_nodes, n_features))
    if var is None:
        variance[:n_samples] = 1
    else:
        variance[:n_samples] = var

    act_level = np.zeros(n_nodes)
    if act is None:
        act_level[:n_samples] = 1
    else:
        act_level[:n_samples] = act

    if 0:  # normalize act levels
        min_a, max_a = min(act), max(act)
        eps = 1e-4
        if max_a != min_a:
            act_level[:n_samples] = (
                act - max_a) / (max_a - min_a) * (1 - eps) + 1 - eps
        else:
            act_level[:n_samples] = 1 - eps

    if (dist_type == 'mixt') or (dist_type == 'mixt_skgmm'):
        act_level = act_level[:n_samples]

    # build moments
    moments_1 = np.zeros(n_nodes)
    if (dist_type == 'uward') or (dist_type == 'uward2'):
        # moments_1: cluster size
        moments_1[:n_samples] = 1
        # moments_2: cluster inertia
        moments_2 = np.zeros((n_nodes, n_features))
        moments_2[:n_samples] = X
    else:  # dist_type == 'mixt'
        # moments_1: label likelihood
        moments_1[:n_samples] = 0
        # moments_2: mixture parameters for each cluster
        moments_2 = np.zeros((n_nodes, n_features * 3 + 3))
        moments_2[:n_samples] = np.hstack((X, X, np.zeros((X.shape[0], n_features)),
                                           np.zeros((X.shape[0], n_features)),
                                           act_level[:, np.newaxis]))

    variance_ini = np.zeros((n_nodes, n_features))
    if var_ini is None:
        variance_ini[:n_samples] = 1
    else:
        variance_ini[:n_samples] = var_ini
    activation_ini = np.zeros(n_nodes)
    if act_ini is None:
        activation_ini[:n_samples] = 1
    else:
        activation_ini[:n_samples] = act_ini

    logger.info('Initialization')
    logger.info('A:')
    logger.info(A)
    logger.info('moments_1: %s', moments_1)
    logger.info('moments_2: %s', moments_2)
    logger.info('coord_row: %s', coord_row)
    logger.info('coord_col: %s', coord_col)
    if (dist_type == 'mixt') or (dist_type == 'mixt_skgmm'):
        logger.info('cluster_masks:')
        logger.info(cluster_masks)

    # Compute initial distances:
    if dist_type == 'uward':
        compute_uward_dist(moments_1, moments_2, coord_row, coord_col,
                           variance_ini, activation_ini, inertia)
    elif dist_type == 'uward2':
        compute_uward_dist2(moments_1, X, act_level,
                            coord_row, coord_col,
                            cluster_masks, inertia)
    elif dist_type == 'mixt':
        compute_mixt_dist(X, act_level, coord_row, coord_col, cluster_masks,
                          moments_1, cov_type, inertia)
    else:  # dist_type == 'mixt_skgmm'
        compute_mixt_dist_skgmm(X, act_level, coord_row, coord_col,
                                cluster_masks, moments_1, cov_type, inertia)

    logger.info('initial inertia:')
    logger.info(inertia)

    inertia = zip(inertia, coord_row, coord_col)
    heapify(inertia)

    # prepare the main fields
    parent = np.arange(n_nodes, dtype=np.int)
    heights = np.zeros(n_nodes)
    used_node = np.ones(n_nodes, dtype=bool)
    children = []
    not_visited = np.empty(n_nodes, dtype=np.int8)

    n_its = n_nodes - n_samples
    history = np.zeros((n_its, n_samples), dtype=np.uint16)
    if save_history:
        history_choices = []
        history_choices_inertia = []
    cur_parc = np.arange(n_samples, dtype=int)

    # recursive merge loop
    for it, k in enumerate(xrange(n_samples, n_nodes)):
        logger.info('iteration: %d', it)
        # identify the merge
        while True:
            inert, i, j = heappop(inertia)
            if used_node[i] and used_node[j]:
                break

        logger.info('group (%d, %d) has min inertia (%f) -> group %d', i, j,
                    inert, k)

        inertia.sort()
        logger.info('remainging inertia heap:')
        logger.info(inertia)

        if save_history:
            choices = []
            choices_inertia = []
            # first choice is kept one:
            choices.append(cluster_masks[i] + cluster_masks[j] * 2)
            choices_inertia.append(inert)
            # other choices:
            for iner, u, v in inertia:
                choices.append(cluster_masks[u] + cluster_masks[v] * 2)
                choices_inertia.append(iner)
            history_choices.append(np.array(choices, dtype=np.int16))
            history_choices_inertia.append(np.array(choices_inertia))

        # print 'i='+str(i)+',j='+str(j)+',k='+str(k)
        parent[i], parent[j], heights[k] = k, k, inert
        children.append([i, j])
        used_node[i] = used_node[j] = False
        new_cluster_mask = np.bitwise_or(cluster_masks[i], cluster_masks[j])
        cluster_masks[k] = new_cluster_mask
        # print 'new_cluster_mask:', new_cluster_mask.astype(int)

        # save history:
        cur_parc[np.where(new_cluster_mask)] = k
        # if save_history:
        history[it] = cur_parc[:]

        # update the moments
        if (dist_type == 'uward'):
            sact = act_level[i] + act_level[j]
            moments_1[k] = moments_1[i] + moments_1[j]
            if sact > 0:
                moments_2[k] = (moments_2[i] * act_level[i] +
                                moments_2[j] * act_level[j]) / sact
                # variance[k] = (variance[i] * act_level[i] + \
                #                variance[j] * act_level[j]) / sact
            else:
                moments_2[k] = moments_2[i] + moments_2[j]
                #variance[k] = variance[i] + variance[j]

            act_level[k] = (act_level[i] + act_level[j])  # /2.
            # act_level[k] = max(act_level[i], act_level[j]) # OK for this
            # shouldnt we update in the same way as moments??
            variance[k] = variance[i] + variance[j]
        elif dist_type == 'uward2':
            m = np.where(new_cluster_mask)
            fm = X[m[0], :]
            am = act_level[m]
            mu0 = fm.mean()
            moments_1[k] = (am * ((fm - mu0) ** 2).T).T.sum()
        elif dist_type == 'mixt':
            m = np.where(new_cluster_mask)
            fm = X[m[0], :]
            am = act_level[m]

            mu0, mu1, v0, v1, l, loglh = informedGMM_MV(fm, am, cov_type)

            moments_1[k] = loglh
            moments_2[k] = np.hstack([mu0, mu1, v0, v1, l])

        elif dist_type == 'mixt_skgmm':
            m = np.where(new_cluster_mask)
            fm = X[m[0], :]
            am = act_level[m]

            #g0 = GMM(n_components=2, covariance_type='spherical')
            # g0.fit(fm)

            g = GMM(n_components=2, covariance_type=cov_type)
            g.fit(fm)

            if cov_type == 'spherical' and not (g.covars_[0, 0] == g.covars_[0, 1]):
                raise Exception('Variances different!!')

            moments_1[k] = g.score(fm).sum()
            moments_2[k] = np.hstack([g.means_[0], g.means_[1],
                                      g.covars_[0, :], g.covars_[1, :],
                                      g.weights_[0]])
        else:
            raise Exception('error unknown dist type: %s' % dist_type)

        # update the structure matrix A and the inertia matrix
        coord_col = []
        not_visited.fill(1)
        not_visited[k] = 0
        _get_parents(A[i], coord_col, parent, not_visited)
        _get_parents(A[j], coord_col, parent, not_visited)
        # List comprehension is faster than a for loop
        [A[l].append(k) for l in coord_col]
        A.append(coord_col)
        coord_col = np.array(coord_col, dtype=np.int)
        coord_row = np.empty_like(coord_col)
        coord_row.fill(k)
        n_additions = len(coord_row)
        ini = np.empty(n_additions, dtype=np.float)

        logger.info('moments_2 of new cluster: %s', moments_2[k])
        logger.info('coord_row: %s', coord_row)
        logger.info('coord_col: %s', coord_col)

        if dist_type == 'uward':
            compute_uward_dist(moments_1, moments_2, coord_row, coord_col,
                               variance, act_level, ini)
        elif dist_type == 'uward2':
            compute_uward_dist2(moments_1, X, act_level,
                                coord_row, coord_col,
                                cluster_masks, ini)
        elif dist_type == 'mixt':
            compute_mixt_dist(X, act_level, coord_row, coord_col,
                              cluster_masks, moments_1, cov_type, ini)

        else:  # dist_type == 'mixt_skgmm'
            compute_mixt_dist_skgmm(X, act_level, coord_row, coord_col,
                                    cluster_masks, moments_1, cov_type, ini)

        # List comprehension is faster than a for loop
        [heappush(inertia, (ini[idx], k, coord_col[idx]))
            for idx in xrange(n_additions)]

    # Separate leaves in children (empty lists up to now)
    n_leaves = n_samples
    children = np.array(children)  # return numpy array for efficient caching

    if save_history:
        # pack history of choices into numpy array
        nmax_choices = max([len(choices) for choices in history_choices])
        print 'size of history_choices:', str((n_its, nmax_choices, n_samples))
        history_choices_a = np.zeros((n_its, nmax_choices, n_samples),
                                     dtype=np.int16) - 1
        #history_choices_a = np.zeros((n_its, 1, n_samples), dtype=np.uint16) - 1
        history_choices_i = np.zeros((n_its, nmax_choices)) - 1
        for it in xrange(n_its):
            nchoices = len(history_choices[it])
            history_choices_a[it, :nchoices, :] = history_choices[it]
            history_choices_i[it, :nchoices] = history_choices_inertia[it]
    else:
        history_choices_a, history_choices_i = None, None

    return children, n_components, n_leaves, parent, heights, moments_2, \
        history, history_choices_a, history_choices_i


###############################################################################
# For non fully-connected graphs

def _fix_connectivity(X, connectivity, n_components, labels):
    """
    Warning: modifies connectivity in place
    """
    for i in range(n_components):
        idx_i = np.where(labels == i)[0]
        Xi = X[idx_i]
        for j in range(i):
            idx_j = np.where(labels == j)[0]
            Xj = X[idx_j]
            D = euclidean_distances(Xi, Xj)
            ii, jj = np.where(D == np.min(D))
            ii = ii[0]
            jj = jj[0]
            connectivity[idx_i[ii], idx_j[jj]] = True
            connectivity[idx_j[jj], idx_i[ii]] = True
    return connectivity

###############################################################################
# Functions for cutting  hierarchical clustering tree


def _hc_cut(n_clusters, children, n_leaves):
    """Function cutting the ward tree for a given number of clusters.

    Parameters
    ----------
    n_clusters : int or ndarray
        The number of clusters to form.

    children : list of pairs. Length of n_nodes
        List of the children of each nodes.
        Leaves have empty list of children and are not stored.

    n_leaves : int
        Number of leaves of the tree.

    Returns
    -------
    labels : array [n_samples]
        cluster labels for each point

    """
    if n_clusters > n_leaves:
        raise ValueError('Cannot extract more clusters than samples: '
                         '%s clusters where given for a tree with %s leaves.'
                         % (n_clusters, n_leaves))
    # In this function, we store nodes as a heap to avoid recomputing
    # the max of the nodes: the first element is always the smallest
    # We use negated indices as heaps work on smallest elements, and we
    # are interested in largest elements
    # children[-1] is the root of the tree
    nodes = [-(max(children[-1]) + 1)]
    for i in range(n_clusters - 1):
        # As we have a heap, nodes[0] is the smallest element
        these_children = children[-nodes[0] - n_leaves]
        # Insert the 2 children and remove the largest node
        heappush(nodes, -these_children[0])
        heappushpop(nodes, -these_children[1])
    label = np.zeros(n_leaves, dtype=np.int)
    for i, node in enumerate(nodes):
        label[_hc_get_descendent(-node, children, n_leaves)] = i
    return label


###############################################################################
# Class for Ward hierarchical clustering

class Ward(BaseEstimator, ClusterMixin):
    """Ward hierarchical clustering: constructs a tree and cuts it.

    Parameters
    ----------
    n_clusters : int or ndarray
        The number of clusters to find.

    connectivity : sparse matrix.
        Connectivity matrix. Defines for each sample the neighboring
        samples following a given structure of the data.
        Default is None, i.e, the hierarchical clustering algorithm is
        unstructured.

    memory : Instance of joblib.Memory or str
        Used to cache the output of the computation of the tree.
        By default, no caching is done. If a string is given, it is the
        path to the caching directory.

    copy : bool
        Copy the connectivity matrix or work inplace.

    n_components : int (optional)
        The number of connected components in the graph defined by the connectivity matrix. If not set, it is estimated.

    compute_full_tree: bool or `auto` (optional)
        Stop early the construction of the tree at n_clusters. This is
        useful to decrease computation time if the number of clusters is
        not small compared to the number of samples. This option is
        useful only when specifying a connectivity matrix. Note also that
        when varying the number of cluster and using caching, it may
        be advantageous to compute the full tree.


    Attributes
    ----------
    children_ : array-like, shape = [n_nodes, 2]
        List of the children of each nodes.  Leaves of the tree do not appear.

    labels_ : array [n_samples]
        cluster labels for each point

    n_leaves_ : int
        Number of leaves in the hierarchical tree.

    n_components_ : sparse matrix.
        The estimated number of connected components in the graph.

    """

    def __init__(self, n_clusters=2, memory=Memory(cachedir=None, verbose=0),
                 connectivity=None, copy=True, n_components=None,
                 compute_full_tree='auto', dist_type='uward', cov_type='spherical',
                 save_history=False):
        self.n_clusters = n_clusters
        self.memory = memory
        self.copy = copy
        self.n_components = n_components
        self.connectivity = connectivity
        self.compute_full_tree = compute_full_tree
        self.dist_type = dist_type
        self.cov_type = cov_type
        self.save_history = save_history

    def fit(self, X, var=None, act=None, var_ini=None, act_ini=None):
        """Fit the hierarchical clustering on the data

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            The samples a.k.a. observations.

        Returns
        -------
        self
        """
        memory = self.memory
        X = array2d(X)
        if isinstance(memory, basestring):
            memory = Memory(cachedir=memory, verbose=0)

        if not self.connectivity is None:
            if not sparse.issparse(self.connectivity):
                raise TypeError("`connectivity` should be a sparse matrix or "
                                "None, got: %r" % type(self.connectivity))

            if (self.connectivity.shape[0] != X.shape[0] or
                    self.connectivity.shape[1] != X.shape[0]):
                raise ValueError("`connectivity` does not have shape "
                                 "(n_samples, n_samples)")

        n_samples = len(X)
        compute_full_tree = self.compute_full_tree
        if self.connectivity is None:
            compute_full_tree = True
        if compute_full_tree == 'auto':
            # Early stopping is likely to give a speed up only for
            # a large number of clusters. The actual threshold
            # implemented here is heuristic
            compute_full_tree = self.n_clusters > max(100, .02 * n_samples)
        n_clusters = self.n_clusters
        if compute_full_tree:
            n_clusters = None

        # Construct the tree
        self.children_, self.n_components_, self.n_leaves_, parents, \
            self.heights, self.moments, self.history, self.history_choices, \
            self.history_choices_inertia = \
            memory.cache(ward_tree)(X, self.connectivity,
                                    n_components=self.n_components,
                                    copy=self.copy, n_clusters=n_clusters,
                                    var=var, act=act,
                                    var_ini=var_ini, act_ini=act_ini,
                                    dist_type=self.dist_type,
                                    cov_type=self.cov_type,
                                    save_history=self.save_history)

        # Cut the tree
        if compute_full_tree:
            self.labels_ = _hc_cut(self.n_clusters, self.children_,
                                   self.n_leaves_)
        else:
            labels = hc_get_heads(parents, copy=False)
            # copy to avoid holding a reference on the original array
            labels = np.copy(labels[:n_samples])
            # Reasign cluster numbers
            self.labels_ = np.searchsorted(np.unique(labels), labels)
        return self


###############################################################################
# Ward-based feature agglomeration

class WardAgglomeration(AgglomerationTransform, Ward):
    """Feature agglomeration based on Ward hierarchical clustering

    Parameters
    ----------
    n_clusters : int or ndarray
        The number of clusters.

    connectivity : sparse matrix
        connectivity matrix. Defines for each feature the neighboring
        features following a given structure of the data.
        Default is None, i.e, the hierarchical agglomeration algorithm is
        unstructured.

    memory : Instance of joblib.Memory or str
        Used to cache the output of the computation of the tree.
        By default, no caching is done. If a string is given, it is the
        path to the caching directory.

    copy : bool
        Copy the connectivity matrix or work inplace.

    n_components : int (optional)
        The number of connected components in the graph defined by the
        connectivity matrix. If not set, it is estimated.

    compute_full_tree: bool or `auto` (optional)
        Stop early the construction of the tree at n_clusters. This is
        useful to decrease computation time if the number of clusters is
        not small compared to the number of samples. This option is
        useful only when specifying a connectivity matrix. Note also that
        when varying the number of cluster and using caching, it may
        be advantageous to compute the full tree.

    variance: array with variances of all samples (default: None)
        Injected in the calculation of the inertia.

    activation: level of activation detected (default: None)
        Used to weight voxels depending on level of activation in inertia computation.
        A non-active voxel will not estimate the HRF correctly,
        so features will not be correct either.

    Attributes
    ----------
    children_ : array-like, shape = [n_nodes, 2]
        List of the children of each nodes.
        Leaves of the tree do not appear.

    labels_ : array [n_samples]
        cluster labels for each point

    n_leaves_ : int
        Number of leaves in the hierarchical tree.

    """

    def fit(self, X, y=None, **params):
        """Fit the hierarchical clustering on the data

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            The data

        Returns
        -------
        self
        """
        return Ward.fit(self, X.T, **params)

# END OF modified hierachical.py from scikit learn

# Feature extraction

import scipy as sp


def calculate_uncertainty(dm, g):
    # beta values. dm: design matrix, g: my glm
    #beta_vars = dict.fromkeys(dm.names)
    uncertainty = np.zeros((g.nvbeta.shape[0], g.s2.shape[0]))
    for ib, bname in enumerate(dm.names):
        # sp.diag(g.nvbeta)[ib]      #variance: diag of cov matrix
        # sig2 = g.s2                #ResMS
        # beta_vars[bname] = sp.diag(g.nvbeta)[ib]*g.s2   #variance for all
        # voxels, condition ib
        uncertainty[ib, :] = sp.diag(g.nvbeta)[ib] * g.s2
    return uncertainty


def hrf_canonical_derivatives(tr, oversampling=2., time_length=25.):
    # Canonical HRF and derivatives
    # build dummy design matrix with a single stim event (1 condition) to get
    # exactly the same hrf, dhrf and d2hrf as in the nipy GLM
    try:
        from collections import OrderedDict
    except ImportError:
        from pyhrf.tools.backports import OrderedDict
    from pyhrf.paradigm import Paradigm
    n_scans = time_length * oversampling
    frametimes = np.linspace(0, (n_scans - 1) * tr / oversampling, n_scans)
    ons = dict([('Pulse', np.array([1]))])
    cnames = ['Pulse']
    onsets = OrderedDict(zip(cnames, [[ons[c]] for c in cnames]))
    paradigm0 = Paradigm(onsets)
    paradigm1 = paradigm0.to_nipy_paradigm()
    from nipy.modalities.fmri import design_matrix as dm
    design_matrix = dm.make_dmtx(frametimes, paradigm1,
                                 # 'spm_time_dispersion',
                                 hrf_model='spm_time_dispersion',
                                 drift_model='Cosine', hfcut=128,
                                 fir_delays=[0])
    return design_matrix.matrix


def compute_hrf(method, my_glm, can, ndelays, i):
    can0 = can[:, 0]
    dcan0 = can[:, 1]
    d2can0 = can[:, 2]
    if method == 'glm':
        hrf = can0[:, np.newaxis] * (my_glm.beta[np.newaxis, i, :])
    if method == 'glm_deriv':
        hrf = can0[:, np.newaxis] * (my_glm.beta[np.newaxis, i * 3 + 0, :]) + \
            dcan0[:, np.newaxis] * (my_glm.beta[np.newaxis, i * 3 + 1, :]) + \
            d2can0[:, np.newaxis] * (my_glm.beta[np.newaxis, i * 3 + 2, :])
    if method == 'glm_deriv5':
        d3can0 = can[:, 3]
        d4can0 = can[:, 4]
        hrf = can0[:, np.newaxis] * (my_glm.beta[np.newaxis, i * 5 + 0, :]) + \
            dcan0[:, np.newaxis] * (my_glm.beta[np.newaxis, i * 5 + 1, :]) + \
            d2can0[:, np.newaxis] * (my_glm.beta[np.newaxis, i * 5 + 2, :]) + \
            d3can0[:, np.newaxis] * (my_glm.beta[np.newaxis, i * 5 + 3, :]) + \
            d4can0[:, np.newaxis] * (my_glm.beta[np.newaxis, i * 5 + 4, :])
    if method == 'fir':
        hrf = my_glm.beta[i * ndelays:(i + 1) * ndelays, :]
    return hrf


def squared_error(n, m):  # it needs to enter stim_induced!!!
    error = np.zeros((n.shape[1], 1))
    for i, x in enumerate(error):
        error[i] = np.mean(np.square(m[:, i] - n[:, i]))
    return error


def FWHM(Y):
    import math
    half_max = max(Y) / 2.
    max_ind = np.argmax(Y)
    if max(Y) > 0 and (abs(max(Y)) - abs(min(Y))) > 0:
        max_sign = math.copysign(1., half_max)
        d1 = max_sign * (half_max - np.array(Y[0:max_ind]))
        sd1 = np.array(sorted(d1))
        sd1_idx = np.argsort(d1)
        lval = sd1[sd1 > 0][0]
        li = sd1_idx[sd1 > 0][0]
        rval = sd1[sd1 < 0][-1]
        ri = sd1_idx[sd1 < 0][-1]
        if abs(lval) > abs(rval):
            left_val = rval
            left_idx = ri
        else:
            left_val = lval
            left_idx = li
        d2 = -max_sign * (half_max - np.array(Y[max_ind:]))
        sd2 = np.array(sorted(d2))
        sd2_idx = np.argsort(d2)
        lval = sd2[sd2 > 0][0]
        li = sd2_idx[sd2 > 0][0]
        rval = sd2[sd2 < 0][-1]
        ri = sd2_idx[sd2 < 0][-1]
        if abs(lval) > abs(rval):
            right_val = rval
            right_idx = ri + d1.shape[0]
        else:
            right_val = lval
            right_idx = li + d1.shape[0]
        fwhm = abs(right_idx - left_idx)
    else:
        fwhm = 0.
    return fwhm


def compute_fwhm(F, dt, a=0):
    # full width-at-half-maximum (FWHM):
    F_max = F.max(a)
    fwhm = np.zeros_like(F_max)
    if a == 2:
        for vx in range(F_max.shape[0]):
            for vy in range(F_max.shape[1]):
                fwhm[vx, vy] = FWHM(F[vx, vy, :]) * dt
    else:
        for v, iv in enumerate(F[0, :]):
            fwhm[v] = FWHM(F[:, v]) * dt
    fwhm[np.where((abs(F.max(a)) - abs(F.min(a))) < 0.)] = 0.
    return fwhm.flatten()


from pyhrf.glm import glm_nipy


def GLM_method(name, data0, ncond, dt=.5, time_length=25., ndelays=0):
    if name == 'glm':
        hrf_model = 'Canonical'
        nfeat = 1
    elif name == 'glm_deriv5':
        hrf_model = 'canonical with 5 derivatives'
        nfeat = 5
    elif name == 'fir':
        hrf_model = 'fir'
        nfeat = ndelays
    elif name == 'glm_deriv':
        hrf_model = 'spm_time_dispersion'
        nfeat = 3
    else:
        raise AssertionError("Method not implemented")
    # GLM, features: my_glm.beta
    oversampling = 1. / dt
    fdelays = range(ndelays)  # FIR duration = nrange*TR
    can = hrf_canonical_derivatives(data0.tr, oversampling, time_length)
    [my_glm, design_matrix, c] = glm_nipy(
        data0, hrf_model=hrf_model, fir_delays=fdelays, contrasts={"activation_contrast": "audio"})
    if 0:
        import matplotlib.pyplot as plt
        import pyhrf.plot as plt2
        design_matrix.show()
        plt.savefig('design_matrix.png')
        plt2.autocrop('design_matrix.png')
        plt.close()
    ca = c['activation_contrast']
    Xb = np.dot(design_matrix.matrix, my_glm.beta)   # fit = X*beta
    # output all betas (see pyhrf.glm.glm_nipy_from_files)
    beta_vars = calculate_uncertainty(design_matrix, my_glm)
    hrf = np.zeros((oversampling * time_length, my_glm.beta.shape[1], ncond))
    ttp = np.zeros((my_glm.beta.shape[1], ncond))
    fwhm = np.zeros((my_glm.beta.shape[1], ncond))
    for i in range(ncond):
        hrf[:, :, i] = compute_hrf(name, my_glm, can, ndelays, i)     # HRFs
        ttp[:, i] = my_glm.beta.argmax(0) * dt
        fwhm[:, i] = compute_fwhm(hrf[:, :, i], dt)
    #mse = squared_error(Xb,stim_induced[0:-1:2,:])
    # , mse
    return Xb, my_glm.beta, hrf, design_matrix, beta_vars, ca.pvalue(), ttp, nfeat, fwhm


def generate_features(parcellation, act_labels, feat_levels, noise_var=0.):
    """Generate noisy features with different levels across positions depending on parcellation and activation clusters.

    Parameters
    ----------
    parcellation : np.ndarray of integers in [1, nb_parcels]
        the input parcellation
    act_labels : binary np.ndarray
        define which positions are active (1) and non-active (0)
    feat_levels : dict of (int : (array((n_features,), float), array(n_features), float)) -> (dict of (parcel_idx : (feat_lvl_inact, feat_lvl_act))
        map a parcel labels to feature levels in non-active and active-pos. Eg: `{1: ([1., .5], [10., 15])}` indicates
        that features in parcel 1 have values [1., .5] in non-active positions (2 features per position) and value 10.
        in active-positions
    noise_var : float>0
        variance of additive Gaussian noise

    Returns
    -------
    np.array((n_positions, n_features), float)
        The simulated the features.

    """

    n_features = len(feat_levels[feat_levels.keys()[0]][0])
    n_positions = parcellation.size

    def is_binary(a):
        labels = np.unique(a)
        return (len(labels) == 1 and (labels[0] == 0 or labels[0] == 1) or
                (labels == [0, 1]).all())

    assert is_binary(act_labels)

    features = np.zeros((n_positions, n_features))
    for p in np.unique(parcellation):
        for l in np.unique(act_labels):
            features[np.bitwise_and(parcellation == p, act_labels == l), :] = \
                feat_levels[p][l]

    return features + np.random.randn(n_positions, n_features) * noise_var ** .5


def represent_features(features, labels, ampl, territories, t, fn):
    """
    Generate chart with features represented.

    Args:
        - features: features to be represented
        - labels: territories
        - ampl: amplitude of the positions

    Return:
        features represented in 2D: the size of the spots depends on ampl,
                                    and the color on labels

    """
    import matplotlib.pyplot as plt
    import pyhrf.plot as plt2
    act = (ampl - min(ampl)) / (max(ampl) - min(ampl))  # activation labels
    plt.figure()
    ecolors = ['none', 'black']
    acolors = ['blue', 'red']
    colors = ['blue', 'cyan', 'yellow', 'red',
              'green', 'magenta', 'black', 'white']
    for i in range(len(labels)):
        if t == 1:
            e = ecolors[labels[i]]
            c = colors[territories[i]]
        else:
            e = acolors[labels[i]]
            c = 'none'
        s0 = act[i] * 100
        plt.scatter(features[i, 0], features[i, 1], color=c, s=s0, edgecolor=e)
    plt.xlabel('feature 1')
    plt.ylabel('feature 2')
    plt.savefig(fn)
    plt.close()


def feature_extraction(fmri_data, method, dt=.5, time_length=25., ncond=1):
    """
    fmri_data (pyhrf.core.FmriData): single ROI fMRI data
    """
    import matplotlib.pyplot as plt
    import pyhrf.plot as plt2
    Xb, f, hrf, dm, beta_vars, pval, ttp, nfeat, fwhm = GLM_method(method, fmri_data, ncond,
                                                                   dt, time_length)
    amplitude0 = f[0, :]  # / beta_vars[0,:]
    a, A = amplitude0.min(), amplitude0.max()
    ampl_dynrange = (amplitude0 - a) / (A - a)
    # For more than 1 cond we sum the beta 0 of the conditions:
    # amplitude = np.zeros((f.shape[1],1))
    # for i in range(ncond):
    #     amplitude += f[i*nfeat,:]
    features = np.zeros((f.shape[1], nfeat - 1))
    uncertainty = np.zeros_like(features)
    # TODO: fix for multiple
    for i, v in enumerate(np.arange(1, nfeat * ncond, 1)):
                                                      # conditions
        features[:, i] = f[v, :]  # / beta_vars[v,:]
        if 1:  # normalize features
            m, M = features[:, i].min(), features[:, i].max()
            features[:, i] = (features[:, i] - M) / (M - m) + 1

        uncertainty[:, i] = beta_vars[v, :]
    #features[:,nfeat-1] = ttp[:,cond]

    return ampl_dynrange, pval, features, uncertainty


# Parcellation

import pyhrf.graph as pgraph




def spatial_ward_sk(features, graph, nb_clusters=0):
    connectivity2 = pgraph.graph_to_sparse_matrix(graph)
    ward = WardSK(n_clusters=nb_clusters,
                  connectivity=connectivity2).fit(features)
    ward.labels_ += 1
    return ward


def spatial_ward(features, graph, nb_clusters=0):
    # Spatial Ward parcellation. Returns distance and parcels
    connectivity2 = pgraph.graph_to_sparse_matrix(graph)
    ward = Ward(n_clusters=nb_clusters, connectivity=connectivity2).fit(
        features)  # control transpose
    ward.labels_ += 1
    ward.features = features
    ward.graph = graph

    return ward


def spatial_ward_with_uncertainty(features, graph, variance, activation, var_ini=None, act_ini=None, nb_clusters=0,
                                  dist_type='uward', cov_type='spherical', save_history=False):
    """Parcellation the given features with the spatial Ward algorithm, taking into account uncertainty on features
    (variance) and activation level:

    - the greater the variance of a given sample, the lower its importance in the distance.
    - the lower the activation level of a given sample, the lower its distance to any other sample.

    Parameters
    ----------
    feature : np.ndarray
        observations to parcellate - size: (nsamples, nfeatures)
    graph : list of (list of PositionIndex)
        spatial dependency between positions
    variance : np.ndarray
        variance of features - size: (nsamples, nfeatures)
    activation : np.ndarray
        activation level associated with observation.
    size : int
        n samples
    var_ini
    act_ini:
    nb_clusters : int
        number of clusters
    dist_type : str
        `ward` | `mixt`

    """
    # Spatial Ward parcellation with uncertainty. Returns distance and parcels
    connectivity2 = pgraph.graph_to_sparse_matrix(graph)
    ward = Ward(n_clusters=nb_clusters, connectivity=connectivity2,
                dist_type=dist_type, cov_type=cov_type, save_history=save_history)
    # control transpose
    ward.fit(features, variance, activation, var_ini, act_ini)
    ward.features = features
    ward.variance = variance
    ward.activation = activation
    ward.graph = graph
    ward.labels_ += 1
    return ward


def parcellation_hemodynamics(fmri_data, feature_extraction_method,
                              parcellation_method, nb_clusters):
    """
    Perform a hemodynamic-driven parcellation on masked fMRI data

    Args:
        - fmri_data (pyhrf.core.FmriData): input fMRI data
        - feature_extraction_method (str): one of
          'glm_hderiv', 'glm_hdisp' ...
        - parcellation_method (str): one of
          'spatial_ward', 'spatial_ward_uncertainty', ...

    Return:
         parcellation array (numpy array of integers) with flatten
         spatial axes

    Examples #TODO
    """
    fmri_data.build_graphs()  # ensures that graphs are built
    roi_ids = np.unique(fmri_data.roi_ids_in_mask)
    if len(roi_ids) == 0:
        # glm
        amplitude, pvalues, features, uncertainty = feature_extraction(fmri_data,
                                                                       feature_extraction_method)
        # ncond=1, cond=0, dt=.5, time_length=25.)

        # parcellation process
        #graph = pgraph.graph_from_lattice(np.ones((5,5)), pgraph.kerMask2D_4n)
        if parcellation_method == 'spatial_ward':
            parcellation = spatial_ward(features, graph, nb_clusters)
        else:
            parcellation = spatial_ward_with_uncertainty(features, graph,
                                                         uncertainty, 1. - pvalues, nb_clusters)

    else:  # parcellate each ROI separately
        nb_voxels_all = fmri_data.nb_voxels_in_mask
        parcellation = np.zeros(fmri_data.nb_voxels_all, dtype=int)
        for rfd in fmri_data.roi_split():
            # multiply nb_clusters by the fraction of the roi size
            nb_clusters_roi = round(nb_clusters * rfd.nb_voxels_in_mask /
                                    nb_voxels_all)
            p_roi = parcellation_hemodynamics(rfd, feature_extraction_method,
                                              parcellation_method,
                                              nb_clusters_roi)
            parcellation += p_roi + parcellation.max()

    return parcellation


from pyhrf.parcellation import parcellation_dist


def assert_parcellation_equal(p1, p2, mask=None, tol=0, tol_pos=None):
    pdist = parcellation_dist(p1, p2, mask)

    if (tol_pos is not None and (tol_pos[np.where(pdist[1] == 0)] != 1).any()) or \
            (tol_pos is None and pdist[0] > tol):
        msg = 'Parcellation are not equal. %d differing positions:\n' \
            % pdist[0]
        msg += str(pdist[1])
        raise AssertionError(msg)


def align_parcellation(p1, p2, mask=None):
    """
    Align two parcellation p1 and p2 as the minimum
    number of positions to remove in order to obtain equal partitions.
    Return:
        (p2 aligned to p1)
    """
    assert np.issubdtype(p1.dtype, np.int)
    assert np.issubdtype(p2.dtype, np.int)
    from munkres import Munkres
    from pyhrf.cparcellation import compute_intersection_matrix
    if mask is None:
        mask = (p1 != 0)
    m = np.where(mask)
    logger.debug('Nb pos inside mask: %d', len(m[0]))
    fp1 = p1[m].astype(np.int32)
    fp2 = p2[m].astype(np.int32)
    cost_matrix = np.zeros((fp1.max() + 1, fp2.max() + 1), dtype=np.int32)
    logger.debug('Cost matrix : %s', str(cost_matrix.shape))
    compute_intersection_matrix(fp1, fp2, cost_matrix)
    # discard 0-labelled parcels (background)
    cost_matrix = cost_matrix[1:, 1:]
    # solve the assignement problem:
    indexes = np.array(Munkres().compute((cost_matrix * -1).tolist()))
    p2_aligned = np.zeros_like(p2)
    for i in range(indexes.shape[0]):
        p2_aligned[np.where(p2 == indexes[i, 1] + 1)] = indexes[i, 0] + 1
    return p2_aligned


def mixtp_to_str(mp):
    n_feat = (len(mp) - 1) / 2
    return 'mu0: %s, mu1: %s, a: %1.2f' % (str(mp[:n_feat]), str(mp[n_feat:2 * n_feat]),
                                           mp[-1])

from pyhrf.ndarray import xndarray, MRI3Daxes
import os.path as op


def ward_tree_save(tree, output_dir, mask):

    def expand_and_save(x, anames, adoms, fn):
        prefix = 'parcellation_%s_' % tree.dist_type
        fn = add_prefix(op.join(output_dir, fn), prefix)
        c = xndarray(x, anames, adoms)
        if 'voxel' in anames:
            c.expand(mask, 'voxel', MRI3Daxes[:mask.ndim]).save(fn)
        else:
            c.save(fn)

    # save inputs:
    expand_and_save(tree.features, ['voxel', 'feature'], {}, 'features.nii')
    if hasattr(tree, 'variance'):
        expand_and_save(
            tree.variance, ['voxel', 'feature'], {}, 'variances.nii')
        expand_and_save(tree.activation, ['voxel'], {}, 'activations.nii')

    # save history
    expand_and_save(tree.history, ['iteration', 'voxel'], {}, 'history.nii')
    if tree.history_choices is not None:
        expand_and_save(tree.history_choices, ['iteration', 'candidate', 'voxel'],
                        {}, 'choice_history.nii')
        expand_and_save(tree.history_choices_inertia, ['iteration', 'candidate'],
                        {}, 'choice_history_inertia.nii')

    # save moments:
    expand_and_save(tree.moments, ['iteration', 'moment'], {}, 'moments.nii')

    # save final output:
    expand_and_save(tree.labels_, ['voxel'], {}, 'labels.nii')


def render_ward_tree(tree, fig_fn, leave_colors=None):

    leave_labels = ['%d: f%s - v%s - a%1.2f' % (i, str(f), str(v), a)
                    for i, (f, v, a) in enumerate(zip(tree.features,
                                                      tree.variance,
                                                      tree.activation))]

    n_leaves = len(leave_labels)
    leave_colors = leave_colors or ["black"] * n_leaves

    import pygraphviz as pgv
    g = pgv.AGraph(directed=True)

    def format_item(i):
        if i < n_leaves:
            return '%s' % leave_labels[i]
        else:  # group node
            return '%d' % i

    # add spatial associations as undirected groups:
    for l, ln in enumerate(tree.graph):
        for n in ln:
            if n > l:
                g.add_edge(format_item(l), format_item(n), dir="none")

    for l in xrange(n_leaves):
        n = g.get_node(format_item(l))
        n.attr['shape'] = 'box'
        n.attr['color'] = leave_colors[l]

    # add agglomerative associations as directed groups:
    for ic, c in enumerate(tree.children_):
        g.add_edge(format_item(c[0]), format_item(n_leaves + ic), color="blue")
        g.add_edge(format_item(c[1]), format_item(n_leaves + ic), color="blue")

    for l in xrange(n_leaves, len(tree.heights)):
        n = g.get_node(format_item(l))
        n.attr['color'] = 'blue'

    # render group info in another component:
    if 1:
        for ic in xrange(len(tree.children_)):
            if ic > 0:
                if tree.dist_type == 'uward':
                    smom0 = str(tree.moments[n_leaves + ic - 1])
                    smom1 = str(tree.moments[n_leaves + ic])
                else:  # dist_type == mixt
                    smom0 = mixtp_to_str(
                        tuple(tree.moments[n_leaves + ic - 1]))
                    smom1 = mixtp_to_str(tuple(tree.moments[n_leaves + ic]))
                g.add_edge(format_item(n_leaves + ic - 1) + ", iner=%1.3f, mom=%s"
                           % (tree.heights[n_leaves + ic - 1], smom0),
                           format_item(n_leaves + ic) + ", iner=%1.3f, mom=%s"
                           % (tree.heights[n_leaves + ic], smom1))

    g.layout('dot')
    g.draw(fig_fn)



# -*- coding: utf-8 -*-

import logging

from time import time
from collections import defaultdict

import numpy
import numpy as np
import scipy.stats

from numpy import (array, int32, float32, fix, sqrt, eye, dot, linalg, kron,
                   newaxis, log, array2string)
from numpy import zeros

import pyhrf

from pyhrf.tools import array_summary
from pyhrf import xmlio
from pyhrf.tools import format_duration
from pyhrf.ndarray import xndarray, stack_cuboids
from pyhrf.boldsynth.hrf import getCanoHRF


logger = logging.getLogger(__name__)


def init_dict():
    return defaultdict(list)


class RFIREstim(xmlio.XmlInitable):
    """
    Class handling the estimation of HRFs from fMRI data.
    Analysis is voxel-wise and can
    be multissession (heteroscedastic noise and session dependent drift).
    Simultaneous analysis of several conditions is treated.
    One HRF is considered at each voxel.
    """

    parametersComments = {
        'hrf_nb_coeffs': 'Number of values in the discrete HRF. '
        'Discretization is homogeneous HRF time length is then: '
        ' nb_hrf_coeffs * hrf_dt ',
        'hrf_dt': 'Required HRF temporal resolution',
        'drift_type': 'Basis type in the drift model. Either "cosine" or "poly"',
    }

    if pyhrf.__usemode__ == pyhrf.ENDUSER:
        default_stop_crit1 = 0.0001
        default_stop_crit2 = 0.00001
        default_nb_its = 500
        parametersToShow = ['hrf_nb_coeffs', 'hrf_dt', 'drift_type',
                            'nb_iterations']
    else:
        default_stop_crit1 = 0.005
        default_stop_crit2 = 0.0005
        default_nb_its = None

    def __init__(self, hrf_nb_coeffs=42, hrf_dt=0.6, drift_type='cosine',
                 stop_crit1=default_stop_crit1, stop_crit2=default_stop_crit2,
                 nb_its_max=5, nb_iterations=default_nb_its, nb_its_min=1,
                 average_bold=False, taum=0.01, lambda_reg=100., fixed_taum=False,
                 discarded_scan_indexes=None, output_fit=False):
        """
           'discarded_scan_indexes' : None if no subsampling done, else give position that were removed after temporal subsampling as a 2d numpy array

        """

        xmlio.XmlInitable.__init__(self)

        self.K = hrf_nb_coeffs
        self.OrthoBtype = drift_type
        self.DeltaT = hrf_dt
        self.nbItMax = nb_its_max
        self.nbItMin = nb_its_min
        self.nbIt = nb_iterations
        self.emStop1 = stop_crit1
        self.emStop2 = stop_crit2

        self.save_history = False

        self.compute_pct_change = False

        self.avg_bold = False

        self.fixed_taum = fixed_taum
        self.taum_init = taum
        self.lambda_reg = lambda_reg

        self.pos_removed = discarded_scan_indexes or np.array(([0]))
        self.output_fit = output_fit

    def linkToData(self, data):

        self.M = data.nbConditions
        dp = data.paradigm
        self.condition_names = dp.stimOnsets.keys()
        self.onsets = dp.stimOnsets.values()
        self.LengthOnsets = [dp.stimDurations[n] for n in dp.stimOnsets.keys()]

        self.nbVoxels = data.bold.shape[1]
        if not self.avg_bold:
            self.bold = data.bold
        else:
            self.bold = data.bold.mean(1)[:, np.newaxis]
            logger.info('BOLD is averaged -> shape: %s', str(self.bold.shape))

        self.sscans = data.sessionsScans
        self.nbSessions = data.nbSessions
        self.I = self.nbSessions
        self.ImagesNb = sum([len(ss) for ss in self.sscans])
        self.TR = data.tr

        self.Ni = array([len(ss) for ss in self.sscans])
        self.OnsetList = [[o[i] for o in self.onsets]
                          for i in xrange(self.nbSessions)]
        self.Qi = zeros(self.nbSessions, dtype=int) + 2

        self.history = defaultdict(init_dict)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def run(self):
        """
        function to launch the analysis
        """
        logger.info('Starting voxel-wise HRF estimation ...')
        logger.info('nvox=%d, ncond=%d, nscans=%d', self.nbVoxels, self.M,
                    self.ImagesNb)
        # initialization of the matrices that will store all voxel resuls
        logger.info("Init storage ...")
        self.InitStorageMat()
        logger.info("Compute onset matrix ...")
        self.Compute_onset_matrix3()

        self.stop_iterations = np.zeros(self.bold.shape[1], dtype=int)

        # voxelwise analysis. This loop handles the currently analyzed voxels.
        for POI in xrange(self.bold.shape[1]):  # POI = point of interest
            t0 = time()
            logger.info("Point %s / %s", str(POI), str(self.nbVoxels))
            self.ReadPointOfInterestData(POI)

            # initialize with zeros or ones all matrix and vector used in
            # the class
            logger.info("Init matrix and vectors ...")
            self.InitMatrixAndVectors(POI)

            # compute onset matrix

            # compute low frequency basis
            logger.info('build low freq mat ...')
            self.buildLowFreqMat()

            # precompute usefull data
            logger.info('Compute inv R ...')
            self.Compute_INV_R_and_R_and_DET_R()

            # solve find response functions and hyperparameters
            logger.info('EM solver ...')
            self.EM_solver(POI)

            # store current results in matrices initialized in 'InitStoringMat'
            logger.info('Store res ...')
            self.StoreRes(POI)
            logger.info("Done in %s", format_duration(time() - t0))

        self.clean_memory()

        logger.info('Nb of iterations to reach stop crit: %s',
                    array_summary(self.stop_iterations))

    def clean_memory(self):
        """ Clean all objects that are useless for outputs
        """
        pass
        # del self.Sigma
        # del self.X
        # del self.R
        # del self.InvSigma
        # del self.InvR
        # del self.h

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def InitStorageMat(self):
        """Initialization of the matrices that will store all voxel results.

        Notes
        -----
        Input signals must have been read (in ReadRealSignal)

        """
        npos = self.bold.shape[1]
        self.Pvalues = -0.1 * numpy.ones((self.M, npos), dtype=float)
        self.ResponseFunctionsEvaluated = numpy.zeros(
            (self.M, self.K + 1, npos), dtype=np.float32)
        if self.compute_pct_change:
            self.ResponseFunctionsEvaluated_PctChgt = numpy.zeros(
                (self.M, self.K + 1, npos), dtype=np.float32)
        # sqrt of the diagonal of the covariance matrix
        self.StdValEvaluated = numpy.zeros(
            (self.M, self.K + 1, npos), dtype=np.float32)
        if self.compute_pct_change:
            # sqrt of the diagonal of the covariance matrix
            self.StdValEvaluated_PctChgt = numpy.zeros(
                (self.M, self.K + 1, npos), dtype=np.float32)
        self.SignalEvaluated = numpy.zeros(
            (self.ImagesNb, npos), dtype=np.float32)

        # orthonormal basis ('P') initialization: self.P[i][n,q] -> n^th value
        # of the q^th function of the orthonormal basis for session i
        self.P = range(self.I)
        for i in xrange(self.I):
            self.P[i] = numpy.zeros((self.Ni[i], self.Qi[i]), dtype=np.float32)

        # drifts coefs ('l') initialization: self.l[i][q] -> q^th coefficient for the the orthormal basis for session i
        # self.l = range(self.I)
        # for i in xrange(self.I):
        #     self.l[i] = numpy.zeros((self.Qi[i]),dtype=float)

        self.l = numpy.zeros((self.I, self.Qi[i], npos), dtype=np.float32)

        # inverse of R
        self.InvR = numpy.zeros((self.K - 1, self.K - 1), dtype=np.float32)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def ReadPointOfInterestData(self, POI):
        """Initialize the parameters for a voxel analysis. The voxel ID is 'POI' in 'ConsideredCoord' initialized in
        'ReadRealSignal'

        Notes
        -----
        Input signals must have been read (in ReadRealSignal)

        """
        # initialization

        #self.Qi = []
        self.y = []

        # analysis paradigm parameters linked to the sessions...
        # the point of interest POI is considered on all sessions
        for i in xrange(self.nbSessions):
            #self.OnsetList.append([o[i] for o in self.onsets])
            self.y.append(self.bold[self.sscans[i], POI])
            # self.Qi.append((int)(2))

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def InitMatrixAndVectors(self, POI):
        """Initialize to zeros: X, y, P, l, h, InvR, Sigma. Initialize to ones: TauM, rb (<-scalar)."""

        logger.info('InitMatrixAndVectors ...')

        # HRFs (h) initialization: self.h[m][k] -> k^{th} HRF estimation value
        # for the condition m  (oversampled time)
        # Warning: since the first and the last values of the HRFs = 0,
        # the nb of HRFs coefs allocated is K-1
        self.h = numpy.zeros((self.M, self.K - 1), dtype=np.float32)
        hcano = getCanoHRF(dt=self.DeltaT,
                           duration=self.K * self.DeltaT)[1]
        self.h[:] = hcano[1:self.K]

        logger.info('self.h:')
        logger.info(self.h)

        for i in xrange(self.I):
            self.P[i][:, :] = 0.

        # drifts coefs ('l') initialization: self.l[i][q] -> q^th coefficient for the the orthormal basis for session i
        # for i in xrange(self.I):
        #    self.l[i,:] = 0.
        self.l[:, :, POI] = 0.

        # inverse of R
        self.InvR[:, :] = 0.

        # Sigma -> block matix => self.Sigma[m*SBS:(m+1)*SBS,n*SBS:(n+1)*SBS]]
        # -> (m,n)^th block of Sigma (variance a posteriori)
        SBS = self.K - 1  # Sigma Block Size
        self.Sigma = numpy.zeros(
            (self.M * SBS, self.M * SBS), dtype=np.float32)
        self.InvSigma = numpy.zeros(
            (self.M * SBS, self.M * SBS), dtype=np.float32)

        # TauM initialization: TauM[m] -> specific dynamics for condition m
        self.TauM = numpy.zeros(self.M, dtype=np.float32) + self.taum_init
        self.OldTauM = numpy.ones(self.M, dtype=np.float32) / 100.

        # rb initialization: self.rb[i] -> noise variance for i^th session
        self.rb = numpy.ones(self.I, dtype=np.float32)
        self.Old_rb = numpy.ones(self.I, dtype=np.float32)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def Compute_INV_R_and_R_and_DET_R(self):
        """Both computes `self.InvR` and `self.DetR`

        **Requires:**

        - K-1
        - InvR initialized

        """

        # for compactness
        SBS = self.K - 1

        # computes the 2nd order derivative matrix in the finite differences
        # senses
        D2 = -2 * eye(SBS).astype(np.float32)

        for i in xrange(SBS):
            if i > 0:
                D2[i, i - 1] = 1
            if i < SBS - 1:
                D2[i, i + 1] = 1

        # removes the conditions h[0]=0. and h[-1]=0.
        # D2[0,0]=1.  # !!!
        # D2[0,1]=-2.  # !!!
        # D2[0,2]=1.  # !!!
        # D2[-1,-1]=1.  # !!!
        # D2[-1,-2]=-2.  # !!!
        # D2[-1,-3]=1.  # !!!

        # removes the a priori
        # D2=eye(SBS)  # !!!

        # Computes InvR
        self.InvR = dot(D2.T, D2)

        # Compute R
        self.R = linalg.inv(self.InvR)

        # computes det(InvR)
        det_InvR = linalg.det(self.InvR)

        # computes det(R)
        self.DetR = 1. / det_InvR

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def Compute_onset_matrix3(self):
        """Computes the onset matrix. Each stimulus onset is considered over a period of `LengthOnsets` seconds if
        (`LengthOnsets > DetlaT`) and a time step otherwise.

        **Requires:**

        - X initialized
        - OnsetList
        - TR
        - DeltaT
        - K
        - LengthOnsets

        where `self.X[i][m,n,k]` is such that:

        - session i (\in 0:I-1)
        - condition m (\in 0:M-1)
        - data nb n (\in 0:Ni[i]-1)
        - hrf coef nb k (\in 0:K-2)

        """
        # Onset matrix ('X') initialization: self.X[i][m,n,k] -> binary value of the onset matrix for session i / condition m / data nb n (real time) / hrf coef nb k (oversampled time)
        # Warning: since the first and the last values of the HRFs = 0, the nb
        # of HRFs coefs allocated is K-1
        self.X = []

        # if subsampling done, change Ni to compute design matrix X (gives the
        # real nb of scans in normal case)
        self.pos_removed = np.array(self.pos_removed)
        # print 'pos removed:', self.pos_removed, type(self.pos_removed)
        # print 'Ni:', self.Ni
        if numpy.size(self.pos_removed) > 1:
            nb_removed = numpy.size(self.pos_removed)
            print 'nb of removed pos:', nb_removed
            # print self.sscans
            self.Ni = array([len(ss) + nb_removed for ss in self.sscans])
            # print self.Ni

        for i in xrange(self.I):
            self.X.append(
                numpy.zeros((self.M, self.Ni[i], self.K - 1), dtype=np.float32))

        # print np.array(self.X).shape, self.Ni[0]
        # tests consistency between 'OnsetList' and the size of 'X'
        I = len(self.X)
        M = self.X[0].shape[0]
        K = self.X[0].shape[2] + 1
        Ni = numpy.zeros(I).astype(int32)
        for i in xrange(I):
            Ni[i] = self.X[i].shape[1]

        if K != self.K or I != self.I or M != self.M or sum(abs(Ni - self.Ni)) > 0.01:
            print 'OnsetList is not consitent with X'

        # computes the onset matrix X
        # BLOCK DESIGNED STIMULI
        if (self.LengthOnsets[0][0][0] > self.DeltaT):
            for i in xrange(I):
                for m in xrange(self.X[i].shape[0]):
                    for j in xrange(self.OnsetList[i][m].shape[0]):
                        # j^th onset of [session i / parameter m] is considered
                        CurrentOnset = self.OnsetList[i][m][j]
                        for n in xrange(Ni[i]):
                            for k in xrange(K - 1):
                                if CurrentOnset <= (n * self.TR - (k + 1.0) * self.DeltaT + 0.001) and CurrentOnset > ((n - 1.) * self.TR - (k + 1.) * self.DeltaT - self.LengthOnsets[m][i][j]):
                                    self.X[i][m][n, k] = (
                                        self.DeltaT / self.TR)

        else:  # EVENT DESIGNED STIMULI
            for i in xrange(I):
                for m in xrange(self.X[i].shape[0]):
                    for j in xrange(self.OnsetList[i][m].shape[0]):
                        # j^th onset of [session i / parameter m] is considered
                        CurrentOnset = self.OnsetList[i][m][j]
                        # print 'onset:', CurrentOnset, CurrentOnset.shape
                        for n in xrange(Ni[i]):
                            for k in xrange(K - 1):
                                if CurrentOnset <= (n * self.TR - (k) * self.DeltaT + 0.001) and CurrentOnset > ((n) * self.TR - (k + 1.) * self.DeltaT):
                                    self.X[i][m][n, k] = 1.
                                    # print 'cond ok'
                                    # print '(n)*self.TR-(k+1.)*self.DeltaT:', (n)*self.TR-(k+1.)*self.DeltaT
                                    # print 'n*self.TR-(k)*self.DeltaT+0.001:',
                                    # n*self.TR-(k)*self.DeltaT+0.001

        # Then remove the lines in X corresponding to removed positions
        # assuming same nb of scans per session and samed removed positions
        # over sessions
        if numpy.size(self.pos_removed) > 1:
            pos_kept = np.arange(self.Ni[0])
            # put 0 where positions were removed!
            pos_kept[self.pos_removed] = -1
            pos_kept_final = np.array(
                [pos_kept[i] for i in xrange(self.Ni[0]) if pos_kept[i] > -1])
            print 'pos kept:', pos_kept_final
            X_true = np.array(self.X)[:, :, pos_kept_final, :]
            print 'X_true:', X_true, X_true.shape
            print 'self.X:', self.X, np.array(self.X).shape

            self.X = X_true

            # put Ni again at its true value   with the actual nb of scans
            # after subsampling
            self.Ni = array([len(ss) for ss in self.sscans])

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def buildLowFreqMat(self):
        """Build the low frequency basis matrix P."""
        for i in xrange(self.I):
            if self.OrthoBtype == 'cosine':
                self.P[i] = self.buildCosMat(self.Qi[i], self.TR, self.Ni[i])
            else:
                self.P[i] = self.buildPolyMat(self.Qi[i], self.TR, self.Ni[i])
                # the output columns number is not always as expected
                self.Qi[i] = self.P[i].shape[1]

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def buildPolyMat(self, fctNb, tr, ny):
        """Build a polynomial low frequency basis in P (adapted from `samplerbase.py`)

        Parameters
        ----------
        fctNb
            columns number in the current session
        tr
            the time resolution of the BOLD data (in second)
        ny
            number of data for the current session

        Notes
        -----
        - there may have no constant column in the orthogonal matrix (the algorithm suppose there is one such column)
        - the columns number is not always as expected

        """
        paramLFD = fctNb - \
            1  # order of the orthonormal basis for the drift component for the current session
        regressors = tr * numpy.arange(0, ny)
        timePower = numpy.arange(0, paramLFD + 1, dtype=int)
        regMat = numpy.zeros((len(regressors), paramLFD + 1), dtype=np.float32)
        for v in xrange(paramLFD + 1):
            regMat[:, v] = regressors[:]
        tPowerMat = numpy.matlib.repmat(timePower, ny, 1)
        lfdMat = numpy.power(regMat, tPowerMat)
        lfdMat = numpy.array(scipy.linalg.orth(lfdMat))
        return lfdMat

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def buildCosMat(self, fctNb, tr, ny):
        """Build a cosine low frequency basis in P (adapted from `samplerbase.py`)

        Parameters
        ----------
        fctNb
            columns number in the current session
        tr
            the time resolution of the BOLD data (in second)
        ny
            number of data for the current session
        """
        paramLFD = fix(2 * (ny * tr) / (fctNb - 1.)
                       )  # order of the orthonormal basis for the drift component for the current session / +1 stands for the mean/cst regressor
        n = numpy.arange(0, ny)
        lfdMat = numpy.zeros((ny, fctNb), dtype=np.float32)
        # lfdMat[:,0] is actually lfdMat[:,0]' with matlab norms
        lfdMat[:, 0] = numpy.ones((1, ny), dtype=np.float32) / sqrt(ny)
        samples = 1 + numpy.arange(fctNb - 1, dtype=int)

        for k in samples:
            lfdMat[:, k] = numpy.sqrt(2. / ny) * \
                numpy.cos(numpy.pi * (2. * n + 1.) * k / (2. * ny))
        return lfdMat

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def CptSigma(self):
        """Computes the Sigma at a given iteration.

        `self.Sigma[m*SBS:(m+1)*SBS,n*SBS:(n+1)*SBS]]` -> `(m,n)^th` block of Sigma in session `i`.

        """
        # 0 ) for compactness
        SBS = self.K - 1

        # 1 ) computes InvSigma
        self.InvSigma[:, :] = 0.

        for m in xrange(self.M):
            if not self.fixed_taum:
                self.InvSigma[
                    m * SBS:(m + 1) * SBS, m * SBS:(m + 1) * SBS] = self.InvR / self.TauM[m]
            else:
                self.InvSigma[
                    m * SBS:(m + 1) * SBS, m * SBS:(m + 1) * SBS] = self.InvR * self.lambda_reg

        for i in xrange(self.I):
            for m in xrange(self.M):
                for n in xrange(self.M):
                    if not self.fixed_taum:
                        self.InvSigma[
                            m * SBS:(m + 1) * SBS, n * SBS:(n + 1) * SBS] += dot(self.X[i][m].T, self.X[i][n]) / self.rb[i]
                    else:
                        self.InvSigma[
                            m * SBS:(m + 1) * SBS, n * SBS:(n + 1) * SBS] += dot(self.X[i][m].T, self.X[i][n])

        # 2 ) computes Sigma
        # TODO: use system solving instead of matrix inversion
        self.Sigma = linalg.inv(self.InvSigma)

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def CptFctQ(self, CptType):
        r"""Computes the function :math:`Q(\Theta',\tilde{\Theta};y)` at a given iteration

        Notes
        -----
        It requires:

        * All parameters and hyperparameters
        * Sigma
        * InvR
        * CptType = 'K_Km1' or 'K_K'

        """
        # 0 ) Initialisation
        SBS = self.K - 1
        result = 0.

        # 1 ) first part -> \sum_{i=1}^{I} Q_{Y_i|H}
        for i in xrange(self.I):
            result += self.rb[i] * ((float)(self.Ni[i]))

        # 2 ) second part -> Q_{H}
        # 2.1 ) First term
        for m in xrange(self.M):
            result -= ((float)(self.K - 1)) * log(self.TauM[m]) / 2.

        # 2.2 ) second term
        for m in xrange(self.M):
            # 2.2.1 ) ...
            result -= self.TauM[m] * \
                dot(self.h[m], dot(self.R, self.h[m])) / 2.

            # 2.2.2 ) ...
            if self.OrthoBtype == 'cosine':
                result -= ((self.OldTauM[m] * self.TauM[m]
                            * dot(self.R, self.R)).trace()) / 2
            else:
                result -= ((self.TauM[m] * self.TauM[m]
                            * dot(self.R, self.R)).trace()) / 2

        # 2.3 ) third term
        result -= ((float)(self.M) / 2.) * log(self.DetR)

        return result

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def StoreRes(self, POI):
        """Store results computed in the voxel defined in POI.

        Notes
        -----
        The estimation at this voxel must have been performed
        """

        for m in xrange(self.M):
            meanLoc = abs(self.bold[:, POI].mean())
            logger.info('Final HRF cond %d, POI %d:', m, POI)
            logger.info(self.h[m])

            # store evaluated response function (percentage of signal change)
            if self.compute_pct_change:
                self.ResponseFunctionsEvaluated_PctChgt[
                    m, 1:self.K, POI] = self.h[m] * 100 / meanLoc

            # store evaluated response function (without normalization)
            self.ResponseFunctionsEvaluated[m, 1:self.K, POI] = self.h[m]

            # store the P-value
            SBS = self.K - 1
            chi2 = dot(self.h[m], dot(
                self.InvSigma[m * SBS:(m + 1) * SBS, m * SBS:(m + 1) * SBS], self.h[m]))
            Pvalue = 1 - scipy.stats.chi2.cdf(chi2, self.h[m].shape[0])
            self.Pvalues[m, POI] = Pvalue

            # store the sqrt of the diagonal of the covariance matrix
            # (percentage of signal change)
            errors = numpy.zeros(self.K + 1, dtype=np.float32)
            for i in xrange(self.K - 1):
                errors[i + 1] = sqrt(self.Sigma[m * SBS + i][m * SBS + i])
            if self.compute_pct_change:
                self.StdValEvaluated_PctChgt[
                    m, :, POI] = errors * 100 / meanLoc
            self.StdValEvaluated[m, :, POI] = errors

        if self.compute_fit(POI).shape < self.SignalEvaluated[:, POI].shape:
            # there are more than one session
            # print 'More than one session'
            self.SignalEvaluated[
                0:self.compute_fit(POI).shape[0], POI] = self.compute_fit(POI)
        else:
            # only one session
            self.SignalEvaluated[:, POI] = self.compute_fit(POI)

    def compute_fit(self, POI):
        # store the signal evaluated in the first session
        # (drift + convolution between evluated HRF and Onset matrix)
        # print 'self.P[0].shape:', self.P[0].shape
        # fit calculated only for the first session
        EvalSignal = dot(self.P[0], self.l[0, :, POI])
        # print 'EvalSignal.shape:', EvalSignal.shape
        for m in xrange(self.M):
            EvalSignal += dot(self.X[0][m], self.h[m])

        return EvalSignal

    def getOutputs(self):
        outputs = {}

        npos = self.bold.shape[1]

        if self.avg_bold:
            pvals = np.repeat(self.Pvalues, self.nbVoxels, axis=1)

        # outputs['pvalue'] = xndarray(pvals.astype(np.float32),
            # axes_names=['condition','voxel'],
            # axes_domains={'condition':self.condition_names})

        hrf_time_axis = np.arange(self.ResponseFunctionsEvaluated.shape[1]) * \
            self.DeltaT
        if self.compute_pct_change:
            hrf_pct_change = self.ResponseFunctionsEvaluated_PctChgt.astype(
                float32)
            if self.avg_bold:
                hrf_pct_change = np.repeat(hrf_pct_change, self.nbVoxels,
                                           axis=2)
            chpc = xndarray(hrf_pct_change, axes_names=['condition', 'time', 'voxel'],
                            axes_domains={'condition': self.condition_names,
                                          'time': hrf_time_axis})
        # hrf_pct_change_errors = self.StdValEvaluated_PctChgt.astype(float32)
        # chpc_errors = xndarray(hrf_pct_change_errors,
        #                      axes_names=['condition','time','voxel'])

        # c = stack_cuboids([chpc, chpc_errors], axis='error',
        #                   domain=['value','error'], axis_pos='last')

        # outputs['ehrf_prct_bold_change'] = c
            outputs['ehrf_prct_bold_change'] = chpc

        rfe = self.ResponseFunctionsEvaluated
        if self.avg_bold:
            rfe = np.repeat(rfe, self.nbVoxels, axis=2)

        ch = xndarray(rfe.astype(float32),
                      axes_names=['condition', 'time', 'voxel'],
                      axes_domains={'condition': self.condition_names,
                                    'time': hrf_time_axis})

        ch_errors = xndarray(self.StdValEvaluated.astype(float32),
                             axes_names=['condition', 'time', 'voxel'])
        outputs['ehrf_error'] = ch_errors

        # c = stack_cuboids([ch, ch_errors], axis='error',
        #                   domain=['value','error'], axis_pos='last')

        # outputs['ehrf'] = c
        if 0:
            print 'ehrf:',
            print ch.descrip()
            print 'ehrf for cond 0', ch.data[0, :, :].shape
            for ih in xrange(ch.data.shape[2]):
                print array2string(ch.data[0, :, ih])

        outputs['ehrf'] = ch

        ad = {'condition': self.condition_names}
        outputs['ehrf_norm'] = xndarray((rfe ** 2).sum(1) ** .5,
                                        axes_names=['condition', 'voxel'],
                                        axes_domains=ad)

        se = self.SignalEvaluated
        if self.avg_bold:
            se = np.repeat(se, self.nbVoxels, axis=1)

        if self.output_fit:
            cfit = xndarray(se.astype(float32),
                            axes_names=['time', 'voxel'])
            bold = self.bold
            if self.avg_bold:
                bold = np.repeat(bold, self.nbVoxels, axis=1)

            cbold = xndarray(self.bold.astype(np.float32),
                             axes_names=['time', 'voxel'])

            outputs['fit'] = stack_cuboids([cfit, cbold], axis="type",
                                           domain=['fit', 'bold'])

            fit_error = sqrt(
                (self.SignalEvaluated - self.bold) ** 2).astype(float32)
            if self.avg_bold:
                fit_error = np.repeat(fit_error, self.nbVoxels, axis=1)
                outputs['fit_error'] = xndarray(fit_error,
                                                axes_names=['time', 'voxel'])

        # save matrix X
        outputs['matX'] = xndarray(self.X[0].astype(float32),
                                   axes_names=['cond', 'time', 'P'],
                                   value_label='value')

        # print 'self.P[0].transpose().shape:', self.P[0].transpose().shape
        # print 'self.bold.shape:', self.bold.shape
        #l = dot(self.P[0].transpose(), self.bold)

        #fu = (self.SignalEvaluated - dot(self.P[0],l)).astype(float32)
        # if self.avg_bold:
        #fu = np.repeat(fu, self.nbVoxels, axis=1)
        #outputs['fit_undrift'] = xndarray(fu, axes_names=['time','voxel'])

        outputs['drift'] = xndarray(dot(self.P[0], self.l[0]).astype(float32),
                                    axes_names=['time', 'voxel'])

        rmse = sqrt(
            (self.SignalEvaluated - self.bold) ** 2).mean(0).astype(float32)
        if self.avg_bold:
            rmse = np.repeat(rmse, self.nbVoxels, axis=0)
        outputs['rmse'] = xndarray(rmse, axes_names=['voxel'])

        if self.save_history:
            h = self.ResponseFunctionsEvaluated
            print 'h.shape:', h.shape
            sh = (self.nbIt,) + h.shape
            print 'sh:', sh
            h_history = np.zeros(sh, dtype=np.float32)
            for POI in xrange(npos):
                # print 'h_history[:,:,:,POI]:', h_history[:,:,:,POI].shape
                # print 'np.array(self.history["h"][POI]):',
                # np.array(self.history['h'][POI]).shape
                h_history[:, :, :, POI] = np.array(self.history['h'][POI])
            if self.avg_bold:
                h_history = np.repeat(h_history, self.nbVoxels, axis=3)
            outputs['ehrf_history'] = xndarray(h_history,
                                               axes_names=['iteration', 'condition',
                                                           'time', 'voxel'])

        return outputs

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # def SaveStoredResults(self):
    #     """
    #     Save the results stored in the storage matrices
    #     requires:
    #         * the whole estimation must have been performed
    #     """
    #     return
    #     #save results
    #     for m in xrange(self.M):
    #         if self.M>1:
    #             CondID="_Cond" + str(m+1)
    #         else:
    #             CondID=""

    #         PtSaveName= self.OutputID + CondID + "_pvalue.nii"
    #         writeImageWithPynifti(self.Pvalues[m,:],PtSaveName)

    #         PtSaveName= self.OutputID + CondID + "_EvalHRF_PrctChanges.nii"
    #         writeImageWithPynifti(self.ResponseFunctionsEvaluated_PctChgt[m,:],PtSaveName)

    #         PtSaveName= self.OutputID + CondID + "_EvalHRF.nii"
    #         writeImageWithPynifti(self.ResponseFunctionsEvaluated[m,:],PtSaveName)

    #         PtSaveName= self.OutputID + CondID + "_StdValEval_PrctChanges.nii"
    #         writeImageWithPynifti(self.StdValEvaluated_PctChgt[m,:],PtSaveName)

    #         PtSaveName= self.OutputID + CondID + "_StdValEval.nii"
    #         writeImageWithPynifti(self.StdValEvaluated[m,:],PtSaveName)

    #     PtSaveName= self.OutputID + "_OriginalSignal.nii"
    #     writeImageWithPynifti(self.bold[0][:],PtSaveName)

    #     PtSaveName= self.OutputID + "_EstimatedSignal.nii"
    #     writeImageWithPynifti(self.SignalEvaluated[:],PtSaveName)

    def cpt_XSigmaX(self, tempTerm2i, SBS, i):
        for m in xrange(self.M):
            for n in xrange(self.M):
                tempTerm2i += dot(self.X[i][m],
                                  dot(self.Sigma[m * SBS:(m + 1) * SBS,
                                                 n * SBS:(n + 1) * SBS],
                                      self.X[i][n].T))
                # nb_cond**2*(nscans*nb_hrf_coeffs**2 + nscans**2*nb_hrf_coeffs)
                # 10**2 * (125 * 40**2 + 125**2 * 40) ~ 8e7

    #++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def EM_solver(self, POI):
        """
        EM_solver
        requires:
            * everything in the class is supposed initialized
        """
        # initialization
        SBS = self.K - 1
        self.CptSigma()
        FctQ_K_K = 1000000000.0
        iteration = 1
        StopTest1 = 1.0
        StopTest2 = 1.0

        BigVector = numpy.zeros(self.Sigma.shape[1], dtype=float)
        templi = [numpy.zeros(self.Ni[i], dtype=float) for i in xrange(self.I)]
        tempTerm2 = [numpy.zeros((self.X[i][0].shape[0], self.X[i][0].shape[0]),
                                 dtype=np.float32) for i in xrange(self.I)]
        delta_h = 1.
        self.epsilon = 1e-4

        # Main loop.
        def stop_crit(it, StopTest1, StopTest2):
            if self.nbIt is not None:
                return it < self.nbIt
            else:
                return it < self.nbItMin or ((it < self.nbItMax) and
                                             ((StopTest1 > self.emStop1) or
                                              (StopTest2 > self.emStop2)))

        def stop_crit_fixed_taum(it):
            if self.nbIt is not None:
                return it < self.nbIt
            else:
                return it < self.nbItMin or ((it < self.nbItMax) and
                                             delta_h > self.epsilon)

        if self.save_history:
            # Store initial values
            z = np.zeros((self.M, 1))
            self.history['h'][POI].append(np.hstack((z, self.h, z)))
            self.history['l'][POI].append(self.l[:, :, POI].copy())
            self.history['rb'][POI].append(self.rb.copy())
            self.history['tau'][POI].append(self.TauM.copy())
            self.history['fit'][POI].append(self.compute_fit(POI))

        if not self.fixed_taum:
            while stop_crit(iteration, StopTest1, StopTest2):

                if 0:
                    print 'Iteration:', iteration
                    print 'nbItMax:', self.nbItMax
                    print 'nbItMin:', self.nbItMin
                # 1 ) estimation of \hat{h^{MAP}}

                # compute_h_MAP(self.y, self.P, ...)

                BigVector[:] = 0.
                for i in xrange(self.I):
                    temp = self.y[i] - dot(self.P[i], self.l[i, :, POI])
                    for m in xrange(self.M):
                        BigVector[m * SBS:(m + 1) * SBS] += dot(self.X[i][m].T, temp) / \
                            self.rb[i]  # nb_sess*nb_cond*nscans*nb_hrf_coeffs
                        # 1*10*125*40 = 5e4

                # (nb_cond*nb_hrf_coeffs)**3
                BigVector2 = dot(self.Sigma, BigVector)
                #(10*40)**3 = 7e6

                for m in xrange(self.M):
                    logger.info('it %d, h^MAP cond %d:', iteration, m)
                    self.h[m, :] = BigVector2[m * SBS:(m + 1) * SBS]
                    logger.info(self.h[m, :])

                if self.save_history:
                    z = np.zeros((self.M, 1))
                    self.history['h'][POI].append(np.hstack((z, self.h, z)))

                # 2 ) estimation of \hat{l_i}
                for i in xrange(self.I):
                    templi[i][:] = 0.
                    for m in xrange(self.M):
                        templi[i] += dot(self.X[i][m], self.h[m])
                        # nb_sess * nb_cond * nscans * nb_hrf_coeffs ~ 5e4
                    self.l[i, :, POI] = dot(self.P[i].T, self.y[i] - templi[i])
                    # nb_sess * n_preg * nscans

                if self.save_history:
                    self.history['l'][POI].append(self.l[:, :, POI].copy())

                # 3 ) estimation of \hat{rb}...
                for i in xrange(self.I):
                    # 3.1 ) ...first term
                    temp = self.y[i] - dot(self.P[i], self.l[i, :, POI])
                    for m in xrange(self.M):
                        temp -= dot(self.X[i][m], self.h[m])

                    term1 = dot(temp, temp)

                    # 3.2 ) ...second term
                    tempTerm2[i][:] = 0.
                    #compute_XSigmaX(tempTerm2[i], SBS, i)
                    self.cpt_XSigmaX(tempTerm2[i], SBS, i)
                    term2 = tempTerm2[i].trace()

                    # 3.3 ) ...total
                    self.Old_rb[i] = self.rb[i]
                    self.rb[i] = (term1 + term2) / ((float)(self.Ni[i]))

                if self.save_history:
                    self.history['rb'][POI].append(self.rb.copy())

                # ... if rb is the same for all sessions, uncomment these lines...
                # temp=0.
                # for i in xrange(self.I):
                #    temp=self.rb[i]*((float)(self.Ni[i]))
                #
                # self.rb[:]=temp/((float)(self.Ni[:].sum()))

                # 4 ) reestimation of sigma
                self.CptSigma()

                # 5 ) estimation of \TauM
                for m in xrange(self.M):
                    self.OldTauM[m] = self.TauM[m].copy()
                    temp = dot(kron(self.h[m][:, newaxis], self.h[m]) +
                               self.Sigma[
                                   m * SBS:(m + 1) * SBS, m * SBS:(m + 1) * SBS],
                               self.InvR)
                    self.TauM[m] = (temp.trace()) / (self.K - 1)

                if self.save_history:
                    self.history['tau'][POI].append(self.TauM.copy())

                if self.save_history:
                    self.history['fit'][POI].append(self.compute_fit(POI))

                # 6 ) Q function estimation at the current iteration and
                # previous
                FctQ_Km1_Km1 = FctQ_K_K
                FctQ_K_K = self.CptFctQ('K_K')
                FctQ_K_Km1 = self.CptFctQ('K_Km1')

                # 7 ) Stopping criterion
                iteration = iteration + 1

                StopTest1 = linalg.norm(FctQ_K_Km1 - FctQ_Km1_Km1) / \
                    linalg.norm(FctQ_K_Km1)

                StopTest2 = 0.0
                for i in xrange(self.I):
                    tempCalc = sqrt(linalg.norm(self.TauM - self.OldTauM) ** 2.0 +
                                    (self.Old_rb[i] - self.rb[i]) ** 2.0) / \
                        sqrt(linalg.norm(self.TauM) ** 2.0 +
                             self.rb[i] ** 2.0)
                    if StopTest2 < tempCalc:
                        StopTest2 = tempCalc
        else:
            h_prev = np.ones(self.M * SBS)

            while stop_crit_fixed_taum(iteration):

                if 0:
                    print 'Iteration:', iteration
                    print 'nbItMax:', self.nbItMax
                    print 'nbItMin:', self.nbItMin
                # 1 ) estimation of \hat{h^{MAP}}

                # compute_h_MAP(self.y, self.P, ...)

                BigVector[:] = 0.
                for i in xrange(self.I):
                    temp = self.y[i] - dot(self.P[i], self.l[i, :, POI])
                    for m in xrange(self.M):
                        BigVector[
                            m * SBS:(m + 1) * SBS] += dot(self.X[i][m].T, temp)
                        # nb_sess*nb_cond*nscans*nb_hrf_coeffs
                        # 1*10*125*40 = 5e4

                # (nb_cond*nb_hrf_coeffs)**3
                BigVector2 = dot(self.Sigma, BigVector)
                #(10*40)**3 = 7e6

                for m in xrange(self.M):
                    logger.info('it %d, h^MAP cond %d:', iteration, m)
                    self.h[m, :] = BigVector2[m * SBS:(m + 1) * SBS]
                    logger.info(self.h[m, :])

                if self.save_history:
                    z = np.zeros((self.M, 1))
                    self.history['h'][POI].append(np.hstack((z, self.h, z)))

                # 2 ) estimation of \hat{l_i}
                for i in xrange(self.I):
                    templi[i][:] = 0.
                    for m in xrange(self.M):
                        templi[i] += dot(self.X[i][m], self.h[m])
                    self.l[i, :, POI] = dot(self.P[i].T, self.y[i] - templi[i])

                if self.save_history:
                    self.history['l'][POI].append(self.l[:, :, POI].copy())

                if self.save_history:
                    self.history['fit'][POI].append(self.compute_fit(POI))

                # 7 ) Stopping criterion
                iteration = iteration + 1

                delta_h = ((h_prev - BigVector2) ** 2).sum() ** .5
                h_prev = BigVector2.copy()

        self.stop_iterations[POI] = iteration

        logger.info(
            "iteration: %s -> delta_h=%s", str(iteration), str(delta_h))


def rfir(func_data, fir_duration=42, fir_dt=.6, nb_its_max=100, nb_its_min=5, fixed_taum=False, lambda_reg=100.):
    """Fit a Regularized FIR on functional data `func_data`:

    - multisession voxel-based fwd model: y = \sum Xh + Pl + b
    - heteroscedastic noise
    - session dependent drift coefficients
    - one HRF per condition
    - solved by Expectation-Minimization (EM) (iterative scheme)

    **Reference:** "Unsupervised robust non-parametric estimation of the hemodynamic response function for any fMRI
    experiment." Ciuciu, J.-B. Poline, G. Marrelec, J. Idier, Ch. Pallier, and H. Benali. IEEE Trans. Med. Imag.,
    22(10):1235-1251, Oct. 2003.

    Parameters
    ----------
    func_data : pyhrf.core.FmriData
    fir_duration : float
        FIR duration in seconds
    fir_dt : float
        FIR temporal resolution
    fixed_taum : bool
        enable faster (drafter) RFIR version where the HRF variance hyper-parameter is fixed.
    lambda_reg : float
        amount of temporal regularization for the HRF. Only used if `fixed_taum` is true.
    nb_its_min
        minimum number of iterations for the EM
    nb_its_max
        maximum number of iterations for the EM

    Returns
    -------
    dict of xndarray instances

    """
    rfir_estimator = RFIREstim(hrf_nb_coeffs=int(np.round(fir_duration / fir_dt)),
                               hrf_dt=fir_dt, nb_its_max=nb_its_max,
                               nb_its_min=nb_its_min, fixed_taum=fixed_taum,
                               lambda_reg=lambda_reg)
    rfir_estimator.linkToData(func_data)
    rfir_estimator.run()
    outputs = rfir_estimator.getOutputs()
    to_return = {'fir': outputs["ehrf"], 'fir_error': outputs["ehrf_error"],
                 'drift': outputs["drift"]}
    if outputs.has_key('fit'):
        to_return['fit'] = outputs['fits']
    return to_return



# -*- coding: utf-8 -*-

import unittest
import shutil

import pyhrf
import pyhrf.boldsynth.scenarios as simu

from pyhrf.rfir import rfir


class RFIRTest(unittest.TestCase):
    """
    Test the Regularized FIR (RFIR)-based methods implemented in pyhrf.rfir
    """

    def setUp(self):
        self.tmp_dir = pyhrf.get_tmp_path()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_rfir_on_small_simulation(self):
        """ Check if pyhrf.rfir runs properly and that returned outputs
        contains the expected items """
        fdata = simu.create_small_bold_simulation()
        outputs = rfir(fdata, nb_its_max=2)

        assert isinstance(outputs, dict)
        for k in ["fir", "fir_error", "drift"]:
            assert outputs.has_key(k)

        # TODO: test shape consistency



# -*- coding: utf-8 -*-


import os
import gzip as _gzip

__all__ = ['gunzip', 'gzip_file']


def gunzip(gzFileName, outFileName=None):
    """
    Gunzip the file 'gzFileName' and store the results in 'outFileName'
    """
    if outFileName is None:
        outFileName = os.path.splitext(gzFileName)[0]
    fgz = _gzip.open(str(gzFileName),'rb')
    content = fgz.readlines()
    fgz.close()
    fOut = open(outFileName, 'w')
    fOut.writelines(content)
    fOut.close()
    
def gzip_file(fileName, outFileName=None):
    if outFileName is None:
        outFileName = fileName + '.gz'
    f = open(str(fileName))
    content = f.readlines()
    f.close()
    fOut = _gzip.open(outFileName, 'w')
    fOut.writelines(content)
    fOut.close()



# -*- coding: utf-8 -*-

import numpy as np

from pyhrf.paradigm import Paradigm
from scipy.stats import truncnorm


PHY_PARAMS_FRISTON00 = {
    'model_name': 'Friston00',
    'tau_s': 1 / .8,
    'eps': .5,
    'eps_max': 10.,
    'tau_m': 1.,
    'tau_f': 1 / .4,
    'alpha_w': .2,
    'E0': .8,
    'V0': .02,
    'TE': 0.04,
    'r0': 100,  # 25 at 1.5T, and rO = 25 (B0/1.5)**2
    'vt0': 80.6,  # 40.3 at 1.5T, and vt0 = 40.3 (B0/1.5)
    'e': 0.4,
    'model': 'RBM',
    'linear': True,
    'obata': False,
    'buxton': False
}         # 2 * E_0 - .2

PHY_PARAMS_DONNET06 = {
    'model_name': 'Donnet06',
    'tau_s': 1.54,
    'eps': .5,
    'eps_max': 10.,  # TODO: check this
    'tau_m': 0.98,
    'tau_f': 2.5,
    'alpha_w': .33,
    'E0': .34,
    'V0': 0.05,   # obata 0.04, Friston and others 0.02, Griffeth13 0.05
    'r0': 100,  # 25 at 1.5T, and rO = 25 (B0/1.5)**2
    'vt0': 80.6,  # 40.3 at 1.5T, and vt0 = 40.3 (B0/1.5)
    'e': 1.43,  # 0.4 or 1
    'TE': 0.04,
    'model': 'RBM',
    'linear': True,
    'obata': False,
    'buxton': False
}

PHY_PARAMS_DENEUX06 = {
    'model_name': 'Deneux06',
    'tau_s': .49,
    'eps': .89,        # new alpha
    'eps_max': 10.,
    'tau_m': .27,      # new tau_0
    'tau_f': .19,
    'alpha_w': .63,    # new w
    'E0': .33,
    'V0': .05,
    'TE': 0.04,
    'r0': 100,  # 25 at 1.5T, and rO = 25 (B0/1.5)**2
    'vt0': 80.6,  # 40.3 at 1.5T, and vt0 = 40.3 (B0/1.5)
    'e': 0.4,
    'model': 'RBM',
    'linear': True,
    'obata': False,
    'buxton': False
}         # 2 * E_0 - .2

PHY_PARAMS_HAVLICEK11 = {
    'model_name': 'Havlicek11',
    'tau_s': 1.54,
    'eps': .54,
    'eps_max': 10.,  # TODO: check this
    'tau_m': 0.98,
    'tau_f': 2.63,
    'alpha_w': .34,
    'E0': .32,
    'V0': .04,   # obata 0.04, Friston and others 0.02, Griffeth13 0.05
    'r0': 100,  # 25 at 1.5T, and rO = 25 (B0/1.5)**2
    'vt0': 80.6,  # 40.3 at 1.5T, and vt0 = 40.3 (B0/1.5)
    'e': 1.43,  # 0.4 or 1
    'TE': 0.04,
    'model': 'RBM',
    'linear': True,
    'obata': False,
    'buxton': False
}

PHY_PARAMS_KHALIDOV11 = {
    'model_name': 'Khalidov11',
    'tau_s': 1.54,
    'eps': .54,
    'eps_max': 10.,  # TODO: check this
    'tau_m': 0.98,
    'tau_f': 2.46,
    'alpha_w': .33,
    'E0': .34,
    'V0': 1,   # obata 0.04, Friston and others 0.02, Griffeth13 0.05
    'r0': 100,  # 25 at 1.5T, and rO = 25 (B0/1.5)**2
    'vt0': 80.6,  # 40.3 at 1.5T, and vt0 = 40.3 (B0/1.5)
    'e': 1.43,  # 0.4 or 1
    'TE': 0.018,
    'model': 'RBM',
    'linear': False,
    'obata': False,
    'buxton': False
}

PHY_PARAMS_DUARTE12 = {
    'model_name': 'Duarte12',
    'tau_s': 0.59,
    'eps': 1.,
    'eps_max': 10.,  # TODO: check this
    'tau_m': 0.91,
    'tau_f': 0.4,
    'alpha_w': .32,
    'E0': .34,
    'V0': .02,   # obata 0.04, Friston and others 0.02, Griffeth13 0.05
    'r0': 100,  # 25 at 1.5T, and rO = 25 (B0/1.5)**2
    'vt0': 80.6,  # 40.3 at 1.5T, and vt0 = 40.3 (B0/1.5)
    'e': 1.43,  # 0.4 or 1
    'TE': 0.04,
    'model': 'RBM',
    'linear': True,
    'obata': False,
    'buxton': False
}


def buildOrder1FiniteDiffMatrix_central(size, dt):
    """Returns a toeplitz matrix for central differences to correct for errors on the first and last points (due to the
    fact that there is no rf[-1] or rf[size] to average with):

    - uses the last point to calculate the first and vise-versa
    - this is acceptable bc the rf is assumed to begin & end at steady state (thus the first and last points should
      both be zero)

    """
    from scipy.linalg import toeplitz

    r = np.zeros(size)
    c = np.zeros(size)
    r[1] = .5
    r[size - 1] = -.5
    c[1] = -.5
    c[size - 1] = .5
    return toeplitz(r, c).T / dt  # WARNING!! Modified. Before: (2*dt)


def create_tbg_neural_efficacies(physiological_params, condition_defs, labels):
    """Create neural efficacy from a truncated bi-Gaussian mixture.

    Parameters
    ----------
    physiological_params : dict (<param_name> : <param_value>)
        parameters of the physiological model
    condition_defs : list of pyhrf.Condition
        list of condition definitions. Each item should have the following fields (moments of the mixture):

        - m_act (0<=float<eff_max): mean of activating component
        - v_act (0<float): variance of activating component
        - v_inact (0<float): variance of non-activating component
    labels : np.array((nb_cond, nb_vox), int)
        binary activation states

    Returns
    -------
    np.array(np.array((nb_cond, nb_vox), float))
        the generated neural efficacies

    TODO: settle how to relate brls and prls to neural efficacy
    """

    eff_max = physiological_params['eps_max']
    eff = []
    for ic, c in enumerate(condition_defs):
        labels_c = labels[ic]
        mask_activ = np.where(labels_c)
        eff_c = truncnorm.rvs(0, eff_max, loc=0., scale=c.v_inact ** .5,
                              size=labels_c.size)
        # truncnorm -> loc is mean, scale is std_dev
        eff_c[mask_activ] = truncnorm.rvs(0, eff_max, loc=c.m_act,
                                          scale=c.v_act ** .5,
                                          size=labels_c.sum())

        eff.append(eff_c)
    return np.vstack(eff)


def phy_integrate_euler(phy_params, tstep, stim, epsilon, Y0=None):
    """Integrate the ODFs of the physiological model with the Euler method.

    Parameters
    ----------
    phy_params : dict (<param_name> : <param_value>)
        parameters of the physiological model
    tstep : float
        time step of the integration, in seconds.
    stim : np.array(nb_steps, float)
        stimulation sequence with temporal resolution equal to the time step of the integration
    epsilon : float
        neural efficacy
    Y0 : np.array(4, float) | None
        initial values for the physiological signals. If None: [0, 1,   1, 1.] s  f_in q  v

    Returns
    -------
    np.array((4, nb_steps), float)
        the integrated physiological signals, where indexes of the first axis correspond to:

        - 0 : flow inducing
        - 1 : inflow
        - 2 : HbR
        - 3 : blood volume

    TODO: should the output signals be rescaled wrt their value at rest?
    """

    epsilon = phy_params['eps']  # WARNING!! Added to compute figures
    tau_s = phy_params['tau_s']
    tau_f = phy_params['tau_f']
    tau_m = phy_params['tau_m']
    alpha_w = phy_params['alpha_w']
    E0 = phy_params['E0']

    def cpt_phy_model_deriv(y, s, epsi, dest):
        N, f_in, v, q = y
        if f_in < 0.:
            #raise Exception('Negative f_in (%f) at t=%f' %(f_in, ti))
            # HACK
            print 'Warning: Negative f_in (%f) at t=%f' % (f_in, ti)
            f_in = 1e-4

        dest[0] = epsi * s - (N / tau_s) - ((f_in - 1) / tau_f)  # dNdt
        dest[1] = N  # dfidt
        dest[2] = (1 / tau_m) * (f_in - v**(1 / alpha_w))  # dvdt
        dest[3] = (1 / tau_m) * ((f_in / E0) * (1 - (1 - E0)**(1 / f_in)) -
                                 (q / v) * (v**(1 / alpha_w)))  # dqdt
        return dest

    res = np.zeros((stim.size + 1, 4))
    res[0, :] = Y0 or np.array([0., 1., 1., 1.])
    #res[0, :] = Y0 or np.array([0., 0., 0., 0.]) 

    for ti in xrange(1, stim.size + 1):
        cpt_phy_model_deriv(res[ti - 1], stim[ti - 1], epsilon, dest=res[ti])
        res[ti] *= tstep
        res[ti] += res[ti - 1]

    if 0:
        print 'res shape = ', res.shape

        import matplotlib.pyplot as plt
        plt.figure()
        plt.plot(res[1:, 0], label='N')
        plt.plot(res[1:, 1], label='fi')
        plt.plot(res[1:, 2], label='v')
        plt.plot(res[1:, 3], label='q')
        plt.legend()
        plt.xlabel('points')
        plt.show()

        import nibabel
        img = nibabel.Nifti1Image(res, np.eye(4))
        nibabel.save(img, 'BM_variables.nii')

    return res[1:, :].T


def create_evoked_physio_signals(physiological_params, paradigm, neural_efficacies, dt, integration_step=.05):
    """Generate evoked hemodynamics signals by integrating a physiological model.

    Parameters
    ----------
    physiological_params : dict (<pname (str)> : <pvalue (float)>)
        parameters of the physiological model. In jde.sandbox.physio see PHY_PARAMS_FRISTON00, PHY_PARAMS_FMRII
    paradigm : pyhrf.paradigm.Paradigm
        the experimental paradigm
    neural_efficacies : np.ndarray (nb_conditions, nb_voxels, float)
        neural efficacies involved in flow inducing signal.
    dt : float
        temporal resolution of the output signals, in second
    integration_step : float
        time step used for integration, in second

    Returns
    -------
    np.array((nb_signals, nb_scans, nb_voxels), float)
        All generated signals, indexes of the first axis correspond to:

        - 0: flow inducing
        - 1: inflow
        - 2: blood volume
        - 3: [HbR]

    """
    # TODO: handle multiple conditions
    # -> create input activity signal [0, 0, eff_c1, eff_c1, 0, 0, eff_c2, ...]
    # for now, take only first condition
    first_cond = paradigm.get_stimulus_names()[0]
    stim = paradigm.get_rastered(integration_step)[first_cond][0]
    neural_efficacies = neural_efficacies[0]
    
    # response matrix intialization
    integrated_vars = np.zeros((4, neural_efficacies.shape[0], stim.shape[0]))
    for i, epsilon in enumerate(neural_efficacies):
        integrated_vars[:, i, :] = phy_integrate_euler(physiological_params,
                                                       integration_step, stim,
                                                       epsilon)

    # downsampling:
    nb_scans = paradigm.get_rastered(dt)[first_cond][0].size
    dsf = int(dt / integration_step)
    return np.swapaxes(integrated_vars[:, :, ::dsf][:, :, :nb_scans], 1, 2)
    #return np.swapaxes(integrated_vars[:, :, :][:, :, :], 1, 2)


def create_k_parameters(physiological_params):
    """ Create field strength dependent parameters k1, k2, k3
    """
    # physiological parameters
    V0 = physiological_params['V0']
    E0 = physiological_params['E0']
    TE = physiological_params['TE']
    r0 = physiological_params['r0']
    vt0 = physiological_params['vt0']
    e = physiological_params['e']
    if physiological_params['model'] == 'RBM':  # RBM
        k1 = 4.3 * vt0 * E0 * TE
        k2 = e * r0 * E0 * TE
        k3 = 1 - physiological_params['e']
    elif physiological_params['buxton']:
        k1 = 7 * E0
        k2 = 2
        k3 = 2 * E0 - 0.2
    else:   # CBM
        k1 = (1 - V0) * 4.3 * vt0 * E0 * TE
        k2 = 2. * E0
        k3 = 1 - physiological_params['e']
    physiological_params['k1'] = k1
    physiological_params['k2'] = k2
    physiological_params['k3'] = k3

    return k1, k2, k3


def create_bold_from_hbr_and_cbv(physiological_params, hbr, cbv):
    """
    Compute BOLD signal from HbR and blood volume variations obtained
    by a physiological model
    """

    # physiological parameters
    V0 = physiological_params['V0']
    k1, k2, k3 = create_k_parameters(physiological_params)

    # linear vs non-linear
    if physiological_params['linear']:
        sign = 1.
        if physiological_params['obata']:  # change of sign
            sign = -1
        bold = V0 * ((k1 + k2) * (1 - hbr) + sign * (k3 - k2) * (1 - cbv))
    else:  # non-linear
        bold = V0 * (k1 * (1 - hbr) + k2 * (1 - hbr / cbv) + k3 * (1 - cbv))

    return bold


def create_physio_brf(physiological_params, response_dt=.5,
                      response_duration=25., return_brf_q_v=False):
    """
    Generate a BOLD response function by integrating a physiological model and
    setting its driving input signal to a single impulse.

    Args:
        - physiological_params (dict (<pname (str)> : <pvalue (float)>)):
            parameters of the physiological model.
            In jde.sandbox.physio see PHY_PARAMS_FRISTON00, PHY_PARAMS_FMRII...
        - response_dt (float): temporal resolution of the response, in second
        - response_duration (float): duration of the response, in second

    Return:
        - np.array(nb_time_coeffs, float)
          -> the BRF (normalized)
        - also return brf_not_normalized, q, v when return_prf_q_v=True
          (for error checking of v and q generation in calc_hrfs)
    """

    p = Paradigm({'c': [np.array([0.])]}, [response_duration],
                 {'c': [np.array([1.])]})
    n = np.array([[1.]])
    s, f, v, q = create_evoked_physio_signals(physiological_params, p, n,
                                              response_dt)
    brf = create_bold_from_hbr_and_cbv(physiological_params, q[:, 0], v[:, 0])
    if return_brf_q_v:
        # WARNING!! Added to compute figures
        #return brf / (brf**2).sum()**.5 , q, v, s, f
        return  brf, q, v, s, f
    else:
        return brf #/ (brf**2).sum()**.5


def create_physio_prf(physiological_params, response_dt=.5,
                      response_duration=25., return_prf_q_v=False):
    """
    Generate a perfusion response function by setting the input driving signal
    of the given physiological model with a single impulse.

    Args:
        - physiological_params (dict (<pname (str)> : <pvalue (float)>)):
            parameters of the physiological model.
            In jde.sandbox.physio see PHY_PARAMS_FRISTON00, PHY_PARAMS_FMRII...
        - response_dt (float): temporal resolution of the response, in second
        - response_duration (float): duration of the response, in second

    Return:
        - np.array(nb_time_coeffs, float)
          -> the PRF
        - also return brf_not_normalized, q, v when return_prf_q_v=True
          (for error checking of v and q generation in calc_hrfs)
    """
    p = Paradigm({'c': [np.array([0.])]}, [response_duration],
                 {'c': [np.array([1.])]})  # response_dt to match convention
    # in JDE analysis
    n = np.array([[1.]])
    s, f, v, q = create_evoked_physio_signals(physiological_params, p, n,
                                              response_dt)
    prf = f[:, 0] - f[0, 0]  # remove y-intercept
    if return_prf_q_v:
        #return prf / (prf**2).sum()**.5  # , q, v
        return prf, q, v
    else:
        return prf #/ (prf**2).sum()**.5


def create_omega_prf(primary_brf, dt, phy_params):
    """ create prf from omega and brf
    """
    omega = linear_rf_operator(primary_brf.shape[0], phy_params, dt)
    prf = np.dot(omega, primary_brf)
    prf = prf / (prf**2).sum()**.5
    return prf


def linear_rf_operator(rf_size, phy_params, dt, calculating_brf=False):
    """
    Calculates the linear operator A needed to convert brf to prf & vis-versa
      prf = (A^{-1})brf
      brf = (A)prf

    Inputs:
      - size of the prf and/or brf (assumed to be same)
      - physiological parameters
      - time resolution of data:
      - if you wish to calculate brf (return A), or prf (return inverse of A)

    Outputs:
      - np.array of size (hrf_size,1) linear operator to convert hrfs
    """
    import numpy as np

    tau_m_inv = 1. / phy_params['tau_m']
    alpha_w = phy_params['alpha_w']
    alpha_w_inv = 1. / phy_params['alpha_w']
    E0 = phy_params['E0']
    V0 = phy_params['V0']
    k1, k2, k3 = create_k_parameters(phy_params)
    c = tau_m_inv * (1 + (1 - E0) * np.log(1 - E0) / E0)

    from pyhrf.sandbox.physio import buildOrder1FiniteDiffMatrix_central
    D = buildOrder1FiniteDiffMatrix_central(rf_size, dt)  # numpy matrix
    eye = np.matrix(np.eye(rf_size))  # numpy matrix

    A3 = - tau_m_inv * ((D + (alpha_w_inv * tau_m_inv) * eye).I)
    A4 = - c * (D + tau_m_inv * eye).I + (D + tau_m_inv * eye).I * ((1 - alpha_w)
          * alpha_w_inv * tau_m_inv**2) * (D + alpha_w_inv * tau_m_inv * eye).I
    # A = V0 * ( (k1+k2)*A4 + (k3-k2)* A3 ) # A = h x2^{-1} = Omega^{-1}

    # linear vs non-linear
    if phy_params['linear']:
        sign = 1.
        if phy_params['obata']:  # change of sign
            sign = -1
        # A = h x2^{-1} = Omega^{-1}
        A = V0 * ((k1 + k2) * A4 + sign * (k3 - k2) * A3)
    else:  # non-linear
        A = V0 * \
            (k1 * A4 + k2 *
             (eye - (eye - A4) * np.linalg.inv(eye - A3)) + k3 * A3)

    if (calculating_brf):
        return A.A
    else:  # calculating_prf
        return (A.I).A


def calc_linear_rfs(simu_brf, simu_prf, phy_params, dt, normalized_rfs=True):
    """
    Calculate 'prf given brf' and 'brf given prf' based on the a linearization
    around steady state of the physiological model as described in Friston 2000

    Input:
      - simu_brf, simu_prf: brf and prf from the physiological simulation
                            from which you wish to calculate the respective
                            prf and brf.
                            Assumed to be of size (1,hrf.size)
      - phy_params
      - normalized_rfs: set to True if simu_hrfs are normalized

    Output:
      - calc_brf, calc_prf: np.arrays of shape (hrf.size, 1)
      - q_linear, v_linear: q and v calculated according to the linearized model

    Note:
    These calculations do not account for any rescaling between brf and prf.
    This means the input simu_brf, simu_prf should NOT be rescaled.

    ** Warning**:
      - this function assumes prf.size == brf.size and uses this to build D, I
      - if making modifications:
        calc_brf, calc_prf have a truncation error (due to the finite difference        
        matrix used) on the order of O(dt)^2. If for any reason a hack is later         
        implemented to set the y-intecepts of brf_calc, prf_calc to zero by
        setting the first row of X4, X3 = 0, this will raise a singular matrix
        error in the calculation of calc_prf (due to X.I command), so this error        
        is helpful in this case
    """

    D = buildOrder1FiniteDiffMatrix_central(simu_prf.size, dt)  # numpy matrix
    I = np.matrix(np.eye(simu_prf.size))  # numpy matrix
    # TODO: elimlinate prf.size dependency

    tau_m = phy_params['tau_m']
    # when tau_m=1, singular matrix formed by (D+tau_m_inv*I)
    tau_m_inv = 1. / tau_m
    alpha_w = phy_params['alpha_w']
    alpha_w_inv = 1. / phy_params['alpha_w']
    E0 = phy_params['E0']
    V0 = phy_params['V0']
    k1, k2, k3 = create_k_parameters(phy_params)
    c = tau_m_inv * (1 + (1 - E0) * np.log(1 - E0) / E0)

    # transform to (hrf.size,1) matrix for calcs
    simu_prf = np.matrix(simu_prf).transpose()
    simu_brf = np.matrix(simu_brf).transpose()

    X3 = tau_m_inv * ((D + (alpha_w_inv * tau_m_inv) * I).I)

    X4 = c * (D + tau_m_inv * I).I - (D + tau_m_inv * I).I * ((1 - alpha_w) 
          * alpha_w_inv * tau_m_inv**2) * (D + alpha_w_inv * tau_m_inv * I).I

    # linear vs non-linear
    if phy_params['linear']:
        sign = 1.
        if phy_params['obata']:  # change of sign
            sign = -1
        # A = h x2^{-1} = Omega^{-1}
        X = V0 * ((k1 + k2) * X4 + sign * (k3 - k2) * X3)
    else:  # non-linear
        X = V0 * (k1 * X4 + k2 * (X4 - X3) * np.linalg.inv(I - X3) + k3 * X3)

    # for error checking
    q_linear = 1 - X4 * (-simu_prf)
    v_linear = 1 - X3 * (-simu_prf)

    calc_brf = X * (-simu_prf)

    calc_prf = -X.I * simu_brf

    # convert to np.arrays
    calc_prf = calc_prf.A
    calc_brf = calc_brf.A
    q_linear = q_linear.A
    v_linear = v_linear.A

    if normalized_rfs:
        calc_prf /= (calc_prf**2).sum()**.5
        calc_brf /= (calc_brf**2).sum()**.5

    return calc_brf, calc_prf, q_linear, v_linear



# -*- coding: utf-8 -*-

"""WARNING: DEPRECATED!
This module implements a logging facility."""

import sys
import inspect
import string
import types
import numpy as _np
import logging

logger = logging.getLogger(__name__)


# VERB_NONE = 0
# VERB_SHORT = 1
# VERB_MAIN_STEPS = 2
# VERB_MAIN_FUNC_CALLS = 3
# VERB_MAIN_VAR_VALS = 4
# VERB_DETAIL_VAR_VALS = 5
# VERB_DEBUG = 6

verboseLevels = {
    0: 'no verbose',
    1: 'minimal verbose',
    2: 'main steps',
    3: 'main function calls',
    4: 'main variable values',
    5: 'detailled variable values',
    6: 'debug (everything)',
}


def unprintable(obj):
    """Determine if object can be printed."""
    return inspect.ismethod(obj) or inspect.ismethoddescriptor(obj) or \
        isinstance(obj, (types.BuiltinFunctionType, types.MethodType,
                         types.FunctionType, _np.ufunc, type))


def dict_to_string(dikt, prefix='', numpy_array_format=None,
                   visited=None, exclude=None):
    """Convert dictionnary to displayable string."""
    if numpy_array_format is None:
        numpy_array_format = {'precision': 3, 'max_line_width': 100}
    logger.debug('dict_to_string received: %s\n'
                 '    ->: %s\n'
                 'unprintable: %s', dikt, type(dikt), unprintable(dikt))
    if visited is None:
        visited = []
    if exclude is None:
        exclude = []

    if (id(dikt) in visited) and not (isinstance(dikt, (int, str, float)) or
                                      _np.isscalar(dikt)):
        logger.debug('Already seen: %s %s', dikt, type(dikt))
        return prefix + ' ### \n'

    visited += [id(dikt)]
    space = '  '
    if isinstance(dikt, (str, unicode)) or _np.isscalar(dikt):
        return prefix + repr(dikt) + '\n'
    elif isinstance(dikt, dict):
        sout = prefix + '{\n'
        for k, v in dikt.iteritems():
            if isinstance(k, dict):
                sout += '\n'
                sout += dict_to_string(k, prefix=prefix + space,
                                       numpy_array_format=numpy_array_format,
                                       visited=visited, exclude=exclude)
            else:
                sout += prefix + str(k)
                logger.debug('k: %s', k)
            sout += ' : ' + '\n'
            sout += dict_to_string(v, prefix=prefix + space,
                                   numpy_array_format=numpy_array_format,
                                   visited=visited, exclude=exclude)
            if sout[-1] != '\n':
                sout += '\n'
        sout += prefix + '}'
        if sout[-1] != '\n':
            sout += '\n'
    elif isinstance(dikt, set):
        sout = prefix + str(dikt)
    elif isinstance(dikt, list) or \
            (isinstance(dikt, _np.ndarray) and dikt.dtype == _np.object):
        sout = prefix + '[\n'
        for e in dikt:
            sout += dict_to_string(e, prefix=prefix + space,
                                   numpy_array_format=numpy_array_format,
                                   visited=visited, exclude=exclude)
            sout += prefix + ',\n'
        sout += prefix + ']\n'
    elif isinstance(dikt, tuple):
        sout = prefix + '(\n'
        for e in dikt:
            sout += dict_to_string(e, prefix=prefix + space,
                                   numpy_array_format=numpy_array_format,
                                   visited=visited, exclude=exclude)
            sout += prefix + ',\n'
        sout += prefix + ')\n'
    elif isinstance(dikt, _np.ndarray):
        if dikt.size < 1000:
            sva = _np.array2string(dikt, **numpy_array_format)
            sout = string.join([prefix + s for s in sva.split('\n')], '\n')
        else:
            sout = prefix + '%s -- %1.3g(%1.3g)[%1.3g;%1.3g]' \
                % (str(dikt.shape), dikt.mean(), dikt.std(),
                   dikt.min(), dikt.max())
        if not sout.endswith('\n'):
            sout += '\n'
    else:
        sout = ''
        for attr in dir(dikt):
            attrobj = getattr(dikt, attr)
            if attr[0] != '_' and not unprintable(attrobj) and \
                    attr not in exclude:

                logger.debug('attr: %s\n'
                             '    -> %s', attr, type(attr))
                sout += prefix + '.' + attr + ':' + '\n'
                sout += dict_to_string(getattr(dikt, attr), prefix=prefix + space,
                                       numpy_array_format=numpy_array_format,
                                       visited=visited, exclude=exclude)
    return sout


class Verbose(object):
    """Do not use. Implement a deprecated logging class. Use logging module."""
    default_state = 0
    line_to_be_continued = 1

    def __init__(self, verbosity=0, log=sys.stdout):
        self.verbosity = verbosity
        self.log = log
        self.state = self.default_state

    def set_log(self, log):
        """Set the log file/stream to use."""
        self.log = log

    def set_verbosity(self, verbosity):
        """Define verbosity level."""
        self.verbosity = verbosity

    def __call__(self, verbosity, msg, new_line=True):
        if self.verbosity >= verbosity:
            if isinstance(self.log, file):
                for line in msg.split('\n'):
                    if self.state == self.default_state:
                        if new_line:
                            print >>self.log, '*' * verbosity, line
                            self.state = self.default_state
                        else:
                            print >>self.log, '*' * verbosity, line,
                            self.state = self.line_to_be_continued
                    elif self.state == self.line_to_be_continued:
                        if new_line:
                            print >>self.log, line
                            self.state = self.default_state
                        else:
                            print >>self.log, line,
                            self.state = self.line_to_be_continued
                if new_line:
                    self.log.flush()
            else:
                for line in msg.split('\n'):
                    slog = ('*' * verbosity) + ' ' + line
                    self.log.write(slog)

    def print_ndarray(self, verbosity, ndarray):
        """Displays a representation of an ndarray instance."""
        if self.verbosity >= verbosity:
            if isinstance(ndarray, _np.ndarray):
                self.print_dict(verbosity, ndarray),
            else:
                self.__call__(verbosity, repr(ndarray))

    def print_dict(self, verbosity, dikt, subprefix='', exclude=None):
        """Displays a representation of a dictionnary."""
        if self.verbosity >= verbosity:
            print dict_to_string(dikt, prefix='*' * verbosity + ' ' + subprefix,
                                 exclude=exclude),
